# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timezone
import itertools

import pytest

from swh.indexer.origin_head import get_head_swhid
from swh.indexer.tests.utils import fill_storage
from swh.model.model import (
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
)
from swh.model.swhids import CoreSWHID
from swh.storage.utils import now


@pytest.fixture
def swh_storage_backend_config():
    """In-memory storage, to make tests go faster."""
    return {"cls": "memory"}


SAMPLE_SNAPSHOT = Snapshot(
    branches={
        b"foo": None,
        b"HEAD": SnapshotBranch(
            target_type=SnapshotTargetType.ALIAS,
            target=b"foo",
        ),
    },
)


def _add_snapshot_to_origin(storage, origin_url, visit_type, snapshot):
    storage.origin_add([Origin(url=origin_url)])
    visit = storage.origin_visit_add(
        [
            OriginVisit(
                origin=origin_url,
                date=datetime(2019, 2, 27, tzinfo=timezone.utc),
                type="pypi",
            )
        ]
    )[0]
    storage.snapshot_add([snapshot])
    visit_status = OriginVisitStatus(
        origin=origin_url,
        visit=visit.visit,
        date=now(),
        status="full",
        snapshot=snapshot.id,
    )
    storage.origin_visit_status_add([visit_status])


@pytest.fixture
def storage(swh_storage):
    fill_storage(swh_storage)
    return swh_storage


def test_git(storage):
    origin_url = "https://github.com/SoftwareHeritage/swh-storage"
    assert get_head_swhid(storage, origin_url) == CoreSWHID.from_string(
        "swh:1:rev:384b12006403cce45d6253e38f7bd77dacef726d"
    )


def test_git_partial_snapshot(storage):
    """Checks partial snapshots are ignored."""
    origin_url = "https://github.com/SoftwareHeritage/swh-core"
    storage.origin_add([Origin(url=origin_url)])
    visit = storage.origin_visit_add(
        [
            OriginVisit(
                origin=origin_url,
                date=datetime(2019, 2, 27, tzinfo=timezone.utc),
                type="git",
            )
        ]
    )[0]
    storage.snapshot_add([SAMPLE_SNAPSHOT])
    visit_status = OriginVisitStatus(
        origin=origin_url,
        visit=visit.visit,
        date=now(),
        status="partial",
        snapshot=SAMPLE_SNAPSHOT.id,
    )
    storage.origin_visit_status_add([visit_status])
    assert get_head_swhid(storage, origin_url) is None


def test_vcs_missing_snapshot(storage):
    origin_url = "https://github.com/SoftwareHeritage/swh-indexer"
    storage.origin_add([Origin(url=origin_url)])
    assert get_head_swhid(storage, origin_url) is None


def test_pypi_missing_branch(storage):
    origin_url = "https://pypi.org/project/abcdef/"
    _add_snapshot_to_origin(storage, origin_url, "pypi", SAMPLE_SNAPSHOT)
    assert get_head_swhid(storage, origin_url) is None


@pytest.mark.parametrize(
    "branches_start,branches_middle,branches_end",
    itertools.product([0, 40, 99, 100, 200], [0, 40, 99, 100, 200], [0, 40, 200]),
)
def test_large_snapshot(storage, branches_start, branches_middle, branches_end):
    rev_id = "8ea98e2fea7d9f6546f49ffdeecc1ab4608c8b79"
    snapshot = Snapshot(
        branches=dict(
            [(f"AAAA{i}".encode(), None) for i in range(branches_start)]
            + [
                (
                    b"HEAD",
                    SnapshotBranch(
                        target_type=SnapshotTargetType.ALIAS, target=b"refs/heads/foo"
                    ),
                )
            ]
            + [(f"aaaa{i}".encode(), None) for i in range(branches_middle)]
            + [
                (
                    b"refs/heads/foo",
                    SnapshotBranch(
                        target_type=SnapshotTargetType.REVISION,
                        target=bytes.fromhex(rev_id),
                    ),
                )
            ]
            + [(f"zzzz{i}".encode(), None) for i in range(branches_end)]
        )
    )

    origin_url = "https://example.org/repo.git"
    _add_snapshot_to_origin(storage, origin_url, "git", snapshot)

    assert get_head_swhid(storage, origin_url) == CoreSWHID.from_string(
        "swh:1:rev:8ea98e2fea7d9f6546f49ffdeecc1ab4608c8b79"
    )


def test_large_snapshot_chained_aliases(storage):
    rev_id = "8ea98e2fea7d9f6546f49ffdeecc1ab4608c8b79"
    snapshot = Snapshot(
        branches=dict(
            [(f"AAAA{i}".encode(), None) for i in range(200)]
            + [
                (
                    b"HEAD",
                    SnapshotBranch(
                        target_type=SnapshotTargetType.ALIAS,
                        target=b"refs/heads/alias2",
                    ),
                )
            ]
            + [(f"aaaa{i}".encode(), None) for i in range(200)]
            + [
                (
                    b"refs/heads/alias2",
                    SnapshotBranch(
                        target_type=SnapshotTargetType.ALIAS,
                        target=b"refs/heads/branch",
                    ),
                )
            ]
            + [(f"refs/heads/bbbb{i}".encode(), None) for i in range(200)]
            + [
                (
                    b"refs/heads/branch",
                    SnapshotBranch(
                        target_type=SnapshotTargetType.REVISION,
                        target=bytes.fromhex(rev_id),
                    ),
                )
            ]
        )
    )

    origin_url = "https://example.org/repo.git"
    _add_snapshot_to_origin(storage, origin_url, "git", snapshot)

    assert get_head_swhid(storage, origin_url) == CoreSWHID.from_string(
        "swh:1:rev:8ea98e2fea7d9f6546f49ffdeecc1ab4608c8b79"
    )


@pytest.mark.parametrize(
    "branches_start,branches_end",
    itertools.product([0, 40, 99, 100, 200], [0, 40, 200]),
)
def test_large_snapshot_dangling_alias(storage, branches_start, branches_end):
    snapshot = Snapshot(
        branches=dict(
            [(f"AAAA{i}".encode(), None) for i in range(branches_start)]
            + [
                (
                    b"HEAD",
                    SnapshotBranch(
                        target_type=SnapshotTargetType.ALIAS, target=b"refs/heads/foo"
                    ),
                )
            ]
            + [(f"zzzz{i}".encode(), None) for i in range(branches_end)]
        )
    )

    origin_url = "https://example.org/repo.git"
    _add_snapshot_to_origin(storage, origin_url, "git", snapshot)

    assert get_head_swhid(storage, origin_url) is None


def test_ftp(storage):
    origin_url = "rsync://ftp.gnu.org/gnu/3dldf"
    assert get_head_swhid(storage, origin_url) == CoreSWHID.from_string(
        "swh:1:rev:8ea98e2fea7d9f6546f49ffdeecc1ab4608c8b79"
    )


def test_ftp_missing_snapshot(storage):
    origin_url = "rsync://ftp.gnu.org/gnu/foobar"
    storage.origin_add([Origin(url=origin_url)])
    assert get_head_swhid(storage, origin_url) is None


def test_deposit(storage):
    origin_url = "https://forge.softwareheritage.org/source/jesuisgpl/"
    storage.origin_add([Origin(url=origin_url)])
    assert get_head_swhid(storage, origin_url) == CoreSWHID.from_string(
        "swh:1:rev:e76ea49c9ffbb7f73611087ba6e999b19e5d71eb"
    )


def test_deposit_missing_snapshot(storage):
    origin_url = "https://forge.softwareheritage.org/source/foobar"
    storage.origin_add(
        [
            Origin(
                url=origin_url,
            )
        ]
    )
    assert get_head_swhid(storage, origin_url) is None


def test_pypi(storage):
    origin_url = "https://old-pypi.example.org/project/limnoria/"
    assert get_head_swhid(storage, origin_url) == CoreSWHID.from_string(
        "swh:1:rev:83b9b6c705b125d0fe6dd86b41109dc5fa32f874"
    )

    origin_url = "https://pypi.org/project/limnoria/"
    assert get_head_swhid(storage, origin_url) == CoreSWHID.from_string(
        "swh:1:rel:83b9b6c705b125d0fe6dd86b41109dc5fa32f874"
    )


def test_svn(storage):
    origin_url = "http://0-512-md.googlecode.com/svn/"
    assert get_head_swhid(storage, origin_url) == CoreSWHID.from_string(
        "swh:1:rev:e43f72e12c88abece79a87b8c9ad232e1b773d18"
    )



# Copyright (C) 2015-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.indexer.storage import converters


def test_db_to_mimetype() -> None:
    input_mimetype = {
        "id": b"some-id",
        "tool_id": 10,
        "tool_name": "some-toolname",
        "tool_version": "some-toolversion",
        "tool_configuration": {},
        "encoding": b"ascii",
        "mimetype": b"text/plain",
    }

    expected_mimetype = {
        "id": b"some-id",
        "encoding": b"ascii",
        "mimetype": b"text/plain",
        "tool": {
            "id": 10,
            "name": "some-toolname",
            "version": "some-toolversion",
            "configuration": {},
        },
    }

    actual_mimetype = converters.db_to_mimetype(input_mimetype)

    assert actual_mimetype == expected_mimetype


def test_db_to_fossology_license() -> None:
    input_license = {
        "id": b"some-id",
        "tool_id": 20,
        "tool_name": "nomossa",
        "tool_version": "5.22",
        "tool_configuration": {},
        "license": "GPL2.0",
    }

    expected_license = {
        "id": b"some-id",
        "license": "GPL2.0",
        "tool": {
            "id": 20,
            "name": "nomossa",
            "version": "5.22",
            "configuration": {},
        },
    }

    actual_license = converters.db_to_fossology_license(input_license)

    assert actual_license == expected_license


def test_db_to_metadata() -> None:
    input_metadata = {
        "id": b"some-id",
        "tool_id": 20,
        "tool_name": "some-toolname",
        "tool_version": "some-toolversion",
        "tool_configuration": {},
        "metadata": b"metadata",
    }

    expected_metadata = {
        "id": b"some-id",
        "metadata": b"metadata",
        "tool": {
            "id": 20,
            "name": "some-toolname",
            "version": "some-toolversion",
            "configuration": {},
        },
    }

    actual_metadata = converters.db_to_metadata(input_metadata)

    assert actual_metadata == expected_metadata



# Copyright (C) 2019-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import os
from typing import Callable, Dict, Iterator, List, Optional
import warnings

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import click

from swh.core.cli import CONTEXT_SETTINGS, AliasedGroup
from swh.core.cli import swh as swh_cli_group

logger = logging.getLogger(__name__)


@swh_cli_group.group(
    name="indexer", context_settings=CONTEXT_SETTINGS, cls=AliasedGroup
)
@click.option(
    "--config-file",
    "-C",
    default=None,
    type=click.Path(
        exists=True,
        dir_okay=False,
    ),
    help="Configuration file.",
)
@click.pass_context
def indexer_cli_group(ctx, config_file):
    """Software Heritage Indexer tools.

    The Indexer is used to mine the content of the archive and extract derived
    information from archive source code artifacts.

    """
    from swh.core import config

    ctx.ensure_object(dict)
    if not config_file:
        config_file = os.environ.get("SWH_CONFIG_FILENAME")

    if config_file:
        if not os.path.exists(config_file):
            raise ValueError("%s does not exist" % config_file)
        cfg = config.read(config_file)
    else:
        cfg = {}

    if "indexer.storage" in cfg:
        warnings.warn(
            "The 'indexer.storage' configuration section should be renamed "
            "as 'indexer_storage'",
            DeprecationWarning,
        )
        cfg["indexer_storage"] = cfg.pop("indexer.storage")

    ctx.obj["config"] = cfg


def _get_api(getter, config, config_key, url):
    if url:
        config[config_key] = {"cls": "remote", "url": url}
    elif config_key not in config:
        raise click.ClickException("Missing configuration for {}".format(config_key))
    return getter(**config[config_key])


@indexer_cli_group.group("mapping")
def mapping():
    """Manage Software Heritage Indexer mappings."""
    pass


@mapping.command("list")
def mapping_list():
    """Prints the list of known mappings."""
    from swh.indexer import metadata_dictionary

    mapping_names = [mapping.name for mapping in metadata_dictionary.MAPPINGS.values()]
    mapping_names.sort()
    for mapping_name in mapping_names:
        click.echo(mapping_name)


@mapping.command("list-terms")
@click.option(
    "--exclude-mapping", multiple=True, help="Exclude the given mapping from the output"
)
@click.option(
    "--concise",
    is_flag=True,
    default=False,
    help="Don't print the list of mappings supporting each term.",
)
def mapping_list_terms(concise, exclude_mapping):
    """Prints the list of known CodeMeta terms, and which mappings
    support them."""
    from swh.indexer import metadata_dictionary

    properties = metadata_dictionary.list_terms()
    for property_name, supported_mappings in sorted(properties.items()):
        supported_mappings = {m.name for m in supported_mappings}
        supported_mappings -= set(exclude_mapping)
        if supported_mappings:
            if concise:
                click.echo(property_name)
            else:
                click.echo("{}:".format(property_name))
                click.echo("\t" + ", ".join(sorted(supported_mappings)))


@mapping.command("translate")
@click.argument("mapping-name")
@click.argument("file", type=click.File("rb"))
def mapping_translate(mapping_name, file):
    """Translates file from mapping-name to codemeta format."""
    import json

    from swh.indexer import metadata_dictionary

    mapping_cls = [
        cls for cls in metadata_dictionary.MAPPINGS.values() if cls.name == mapping_name
    ]
    if not mapping_cls:
        raise click.ClickException("Unknown mapping {}".format(mapping_name))
    assert len(mapping_cls) == 1
    mapping_cls = mapping_cls[0]
    mapping = mapping_cls()
    codemeta_doc = mapping.translate(file.read())
    click.echo(json.dumps(codemeta_doc, indent=4))


def list_origins_by_producer(idx_storage, mappings, tool_ids) -> Iterator[str]:
    next_page_token = ""
    limit = 10000
    while next_page_token is not None:
        result = idx_storage.origin_intrinsic_metadata_search_by_producer(
            page_token=next_page_token,
            limit=limit,
            ids_only=True,
            mappings=mappings or None,
            tool_ids=tool_ids or None,
        )
        next_page_token = result.next_page_token
        yield from result.results


@indexer_cli_group.command("journal-client")
@click.argument(
    "indexer",
    type=click.Choice(
        [
            "origin_intrinsic_metadata",
            "extrinsic_metadata",
            "content_mimetype",
            "content_fossology_license",
            "*",
        ]
    ),
    required=True,
)
@click.option(
    "--broker", "brokers", type=str, multiple=True, help="Kafka broker to connect to."
)
@click.option(
    "--prefix", type=str, default=None, help="Prefix of Kafka topic names to read from."
)
@click.option("--group-id", type=str, help="Consumer/group id for reading from Kafka.")
@click.option(
    "--stop-after-objects",
    "-m",
    default=None,
    type=int,
    help="Maximum number of objects to replay. Default is to run forever.",
)
@click.option(
    "--batch-size",
    "-b",
    default=None,
    type=int,
    help="Batch size. Default is 200.",
)
@click.pass_context
def journal_client(
    ctx,
    indexer: Optional[str],
    brokers: List[str],
    prefix: str,
    group_id: str,
    stop_after_objects: Optional[int],
    batch_size: Optional[int],
):
    """Listens for new objects from the SWH Journal, and runs the indexer with
    the name passed as argument

    Passing '*' as indexer name runs all indexers.

    """
    from swh.indexer.indexer import BaseIndexer, ObjectsDict
    from swh.journal.client import get_journal_client

    cfg = ctx.obj["config"]
    journal_cfg = cfg.get("journal_client", cfg.get("journal", {}))

    if brokers:
        journal_cfg["brokers"] = brokers
    if not journal_cfg.get("brokers"):
        raise ValueError("The brokers configuration is mandatory.")

    if prefix:
        journal_cfg["prefix"] = prefix
    if group_id:
        journal_cfg["group_id"] = group_id
    if stop_after_objects:
        journal_cfg["stop_after_objects"] = stop_after_objects
    if batch_size:
        journal_cfg["batch_size"] = batch_size

    object_types = set()
    worker_fns: List[Callable[[ObjectsDict], Dict]] = []

    idx: Optional[BaseIndexer] = None

    if indexer in ("origin_intrinsic_metadata", "*"):
        from swh.indexer.metadata import OriginMetadataIndexer

        object_types.add("origin_visit_status")
        idx = OriginMetadataIndexer()
        idx.catch_exceptions = False  # don't commit offsets if indexation failed
        worker_fns.append(idx.process_journal_objects)

    if indexer in ("extrinsic_metadata", "*"):
        from swh.indexer.metadata import ExtrinsicMetadataIndexer

        object_types.add("raw_extrinsic_metadata")
        idx = ExtrinsicMetadataIndexer()
        idx.catch_exceptions = False  # don't commit offsets if indexation failed
        worker_fns.append(idx.process_journal_objects)

    if indexer in ("content_mimetype", "*"):
        from swh.indexer.mimetype import MimetypeIndexer

        object_types.add("content")
        idx = MimetypeIndexer()
        idx.catch_exceptions = False  # don't commit offsets if indexation failed
        worker_fns.append(idx.process_journal_objects)

    if indexer in ("content_fossology_license", "*"):
        from swh.indexer.fossology_license import FossologyLicenseIndexer

        object_types.add("content")
        idx = FossologyLicenseIndexer()
        idx.catch_exceptions = False  # don't commit offsets if indexation failed
        worker_fns.append(idx.process_journal_objects)

    if not worker_fns:
        raise click.ClickException(f"Unknown indexer: {indexer}")

    if "cls" not in journal_cfg:
        journal_cfg["cls"] = "kafka"

    if (
        journal_cfg.get("object_types")
        and set(journal_cfg["object_types"]) != object_types
    ):
        logger.warning(
            "Overriding configured journal client object types (%s) with %s",
            ", ".join(sorted(journal_cfg["object_types"])),
            ", ".join(sorted(object_types)),
        )

    journal_cfg["object_types"] = list(object_types)

    client = get_journal_client(**journal_cfg)

    def worker_fn(objects: ObjectsDict):
        for fn in worker_fns:
            fn(objects)

    try:
        client.process(worker_fn)
    except KeyboardInterrupt:
        ctx.exit(0)
    else:
        print("Done.")
    finally:
        client.close()


@indexer_cli_group.command("rpc-serve")
@click.argument("config-path", required=True)
@click.option("--host", default="0.0.0.0", help="Host to run the server")
@click.option("--port", default=5007, type=click.INT, help="Binding port of the server")
@click.option(
    "--debug/--nodebug",
    default=True,
    help="Indicates if the server should run in debug mode",
)
def rpc_server(config_path, host, port, debug):
    """Starts a Software Heritage Indexer RPC HTTP server."""
    from swh.indexer.storage.api.server import app, load_and_check_config

    api_cfg = load_and_check_config(config_path, type="any")
    app.config.update(api_cfg)
    app.run(host, port=int(port), debug=bool(debug))


def main():
    return indexer_cli_group(auto_envvar_prefix="SWH_INDEXER")


if __name__ == "__main__":
    main()



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os.path
import re

from rdflib import RDF, BNode, Graph, Literal, URIRef

from swh.indexer.codemeta import _DATA_DIR, read_crosstable
from swh.indexer.namespaces import SCHEMA

from .base import SingleFileIntrinsicMapping, YamlMapping
from .utils import add_map

SPDX = URIRef("https://spdx.org/licenses/")

PUB_TABLE_PATH = os.path.join(_DATA_DIR, "pubspec.csv")

with open(PUB_TABLE_PATH) as fd:
    (CODEMETA_TERMS, PUB_TABLE) = read_crosstable(fd)


def name_to_person(name):
    return {
        "@type": SCHEMA.Person,
        SCHEMA.name: name,
    }


class PubspecMapping(YamlMapping, SingleFileIntrinsicMapping):
    name = "pubspec"
    filename = b"pubspec.yaml"
    mapping = PUB_TABLE["Pubspec"]
    string_fields = [
        "repository",
        "keywords",
        "description",
        "name",
        "issue_tracker",
        "platforms",
        "license",
        # license will only be used with the SPDX Identifier
    ]
    uri_fields = ["homepage"]

    def normalize_license(self, s):
        if isinstance(s, str):
            return SPDX + s

    def _translate_author(self, graph, s):
        name_email_re = re.compile("(?P<name>.*?)( <(?P<email>.*)>)")
        if isinstance(s, str):
            author = BNode()
            graph.add((author, RDF.type, SCHEMA.Person))
            match = name_email_re.search(s)
            if match:
                name = match.group("name")
                email = match.group("email")
                graph.add((author, SCHEMA.email, Literal(email)))
            else:
                name = s

            graph.add((author, SCHEMA.name, Literal(name)))

            return author

    def translate_author(self, graph: Graph, root, s) -> None:
        add_map(graph, root, SCHEMA.author, self._translate_author, [s])

    def translate_authors(self, graph: Graph, root, authors) -> None:
        if isinstance(authors, list):
            add_map(graph, root, SCHEMA.author, self._translate_author, authors)



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


class IndexerStorageAPIError(Exception):
    """Generic error of the indexer storage."""

    pass


class IndexerStorageArgumentException(Exception):
    """Argument passed to an IndexerStorage endpoint is invalid."""

    pass


class DuplicateId(IndexerStorageArgumentException):
    """The same identifier is present more than once."""

    pass



# Copyright (C) 2018-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import ast
import itertools
import re

from rdflib import RDF, BNode, Graph, Literal, URIRef

from swh.indexer.codemeta import CROSSWALK_TABLE
from swh.indexer.namespaces import SCHEMA

from .base import DictMapping, SingleFileIntrinsicMapping
from .utils import add_map

SPDX = URIRef("https://spdx.org/licenses/")


def name_to_person(graph: Graph, name):
    if not isinstance(name, str):
        return None
    author = BNode()
    graph.add((author, RDF.type, SCHEMA.Person))
    graph.add((author, SCHEMA.name, Literal(name)))
    return author


class GemspecMapping(DictMapping, SingleFileIntrinsicMapping):
    name = "gemspec"
    filename = re.compile(rb".*\.gemspec")
    mapping = CROSSWALK_TABLE["Ruby Gem"]
    string_fields = ["name", "version", "description", "summary", "email"]
    uri_fields = ["homepage"]

    _re_spec_new = re.compile(r".*Gem::Specification.new +(do|\{) +\|.*\|.*")
    _re_spec_entry = re.compile(r"\s*\w+\.(?P<key>\w+)\s*=\s*(?P<expr>.*)")

    def translate(self, raw_content):
        try:
            raw_content = raw_content.decode()
        except UnicodeDecodeError:
            self.log.warning("Error unidecoding from %s", self.log_suffix)
            return

        # Skip lines before 'Gem::Specification.new'
        lines = itertools.dropwhile(
            lambda x: not self._re_spec_new.match(x), raw_content.split("\n")
        )

        try:
            next(lines)  # Consume 'Gem::Specification.new'
        except StopIteration:
            self.log.warning("Could not find Gem::Specification in %s", self.log_suffix)
            return

        content_dict = {}
        for line in lines:
            match = self._re_spec_entry.match(line)
            if match:
                value = self.eval_ruby_expression(match.group("expr"))
                if value:
                    content_dict[match.group("key")] = value
        return self._translate_dict(content_dict)

    def eval_ruby_expression(self, expr):
        """Very simple evaluator of Ruby expressions.

        >>> GemspecMapping().eval_ruby_expression('"Foo bar"')
        'Foo bar'
        >>> GemspecMapping().eval_ruby_expression("'Foo bar'")
        'Foo bar'
        >>> GemspecMapping().eval_ruby_expression("['Foo', 'bar']")
        ['Foo', 'bar']
        >>> GemspecMapping().eval_ruby_expression("'Foo bar'.freeze")
        'Foo bar'
        >>> GemspecMapping().eval_ruby_expression( \
                "['Foo'.freeze, 'bar'.freeze]")
        ['Foo', 'bar']
        """

        def evaluator(node):
            if isinstance(node, ast.Str):
                return node.s
            elif isinstance(node, ast.List):
                res = []
                for element in node.elts:
                    val = evaluator(element)
                    if not val:
                        return
                    res.append(val)
                return res

        expr = expr.replace(".freeze", "")
        try:
            # We're parsing Ruby expressions here, but Python's
            # ast.parse works for very simple Ruby expressions
            # (mainly strings delimited with " or ', and lists
            # of such strings).
            tree = ast.parse(expr, mode="eval")
        except (SyntaxError, ValueError):
            return
        if isinstance(tree, ast.Expression):
            return evaluator(tree.body)

    def normalize_license(self, s):
        if isinstance(s, str):
            return SPDX + s

    def normalize_licenses(self, licenses):
        if isinstance(licenses, list):
            return [SPDX + license for license in licenses if isinstance(license, str)]

    def translate_author(self, graph: Graph, root, author):
        if isinstance(author, str):
            add_map(graph, root, SCHEMA.author, name_to_person, [author])

    def translate_authors(self, graph: Graph, root, authors):
        if isinstance(authors, list):
            add_map(graph, root, SCHEMA.author, name_to_person, authors)



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.indexer.metadata_dictionary import MAPPINGS

CONTEXT = [
    "https://doi.org/10.5063/schema/codemeta-2.0",
    {
        "as": "https://www.w3.org/ns/activitystreams#",
        "forge": "https://forgefed.org/ns#",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
    },
]


def test_compute_metadata_none():
    """
    testing content empty content is empty
    should return None
    """
    content = b""

    # None if no metadata was found or an error occurred
    declared_metadata = None
    result = MAPPINGS["GitHubMapping"]().translate(content)
    assert declared_metadata == result


def test_supported_terms():
    terms = MAPPINGS["GitHubMapping"].supported_terms()
    assert {
        "http://schema.org/name",
        "http://schema.org/license",
        "http://schema.org/dateCreated",
        "https://forgefed.org/ns#forks",
        "https://www.w3.org/ns/activitystreams#totalItems",
    } <= terms


def test_compute_metadata_github():
    content = b"""
{
  "id": 80521091,
  "node_id": "MDEwOlJlcG9zaXRvcnk4MDUyMTA5MQ==",
  "name": "swh-indexer",
  "full_name": "SoftwareHeritage/swh-indexer",
  "private": false,
  "owner": {
    "login": "SoftwareHeritage",
    "id": 18555939,
    "node_id": "MDEyOk9yZ2FuaXphdGlvbjE4NTU1OTM5",
    "avatar_url": "https://avatars.githubusercontent.com/u/18555939?v=4",
    "gravatar_id": "",
    "url": "https://api.github.com/users/SoftwareHeritage",
    "type": "Organization",
    "site_admin": false
  },
  "html_url": "https://github.com/SoftwareHeritage/swh-indexer",
  "description": "GitHub mirror of Metadata indexer",
  "fork": false,
  "url": "https://api.github.com/repos/SoftwareHeritage/swh-indexer",
  "created_at": "2017-01-31T13:05:39Z",
  "updated_at": "2022-06-22T08:02:20Z",
  "pushed_at": "2022-06-29T09:01:08Z",
  "archive_url": "https://api.github.com/repos/SoftwareHeritage/swh-indexer/{archive_format}{/ref}",
  "issues_url": "https://api.github.com/repos/SoftwareHeritage/swh-indexer/issues{/number}",
  "git_url": "git://github.com/SoftwareHeritage/swh-indexer.git",
  "ssh_url": "git@github.com:SoftwareHeritage/swh-indexer.git",
  "clone_url": "https://github.com/SoftwareHeritage/swh-indexer.git",
  "svn_url": "https://github.com/SoftwareHeritage/swh-indexer",
  "homepage": "https://forge.softwareheritage.org/source/swh-indexer/",
  "size": 2713,
  "stargazers_count": 13,
  "watchers_count": 12,
  "language": "Python",
  "has_issues": false,
  "has_projects": false,
  "has_downloads": true,
  "has_wiki": false,
  "has_pages": false,
  "forks_count": 1,
  "mirror_url": null,
  "archived": false,
  "disabled": false,
  "open_issues_count": 0,
  "license": {
    "key": "gpl-3.0",
    "name": "GNU General Public License v3.0",
    "spdx_id": "GPL-3.0",
    "url": "https://api.github.com/licenses/gpl-3.0",
    "node_id": "MDc6TGljZW5zZTk="
  },
  "allow_forking": true,
  "is_template": false,
  "web_commit_signoff_required": false,
  "topics": [

  ],
  "visibility": "public",
  "forks": 1,
  "open_issues": 0,
  "watchers": 12,
  "default_branch": "master",
  "temp_clone_token": null,
  "organization": {
    "login": "SoftwareHeritage",
    "id": 18555939,
    "node_id": "MDEyOk9yZ2FuaXphdGlvbjE4NTU1OTM5",
    "avatar_url": "https://avatars.githubusercontent.com/u/18555939?v=4",
    "gravatar_id": "",
    "type": "Organization",
    "site_admin": false
  },
  "network_count": 1,
  "subscribers_count": 6
}

    """  # noqa
    result = MAPPINGS["GitHubMapping"]().translate(content)
    assert result == {
        "@context": CONTEXT,
        "type": "forge:Repository",
        "id": "https://github.com/SoftwareHeritage/swh-indexer",
        "forge:forks": {
            "as:totalItems": {
                "type": "xsd:nonNegativeInteger",
                "@value": "1",
            },
            "type": "as:OrderedCollection",
        },
        "as:likes": {
            "as:totalItems": {
                "type": "xsd:nonNegativeInteger",
                "@value": "13",
            },
            "type": "as:Collection",
        },
        "as:followers": {
            "as:totalItems": {
                "type": "xsd:nonNegativeInteger",
                "@value": "6",
            },
            "type": "as:Collection",
        },
        "license": "https://spdx.org/licenses/GPL-3.0",
        "name": "SoftwareHeritage/swh-indexer",
        "description": "GitHub mirror of Metadata indexer",
        "codeRepository": "https://github.com/SoftwareHeritage/swh-indexer.git",
        "dateCreated": "2017-01-31T13:05:39Z",
        "dateModified": "2022-06-22T08:02:20Z",
        "programmingLanguage": "Python",
    }


def test_github_topics():
    content = b"""
{
  "html_url": "https://github.com/SoftwareHeritage/swh-indexer",
  "topics": [
    "foo",
    "bar"
  ]
}
    """
    result = MAPPINGS["GitHubMapping"]().translate(content)
    assert set(result.pop("keywords", [])) == {"foo", "bar"}, result
    assert result == {
        "@context": CONTEXT,
        "type": "forge:Repository",
        "id": "https://github.com/SoftwareHeritage/swh-indexer",
    }


def test_github_fork():
    content = b"""
{
  "name": "unicode_names2",
  "full_name": "progval/unicode_names2",
  "html_url": "https://github.com/progval/unicode_names2",
  "description": "char <-> Unicode character name (maintained fork of huonw/unicode_names)",
  "parent": {
    "id": 23110520,
    "node_id": "MDEwOlJlcG9zaXRvcnkyMzExMDUyMA==",
    "name": "unicode_names",
    "full_name": "huonw/unicode_names",
    "private": false,
    "html_url": "https://github.com/huonw/unicode_names",
    "description": "char <-> Unicode character name"
  }
}
    """
    result = MAPPINGS["GitHubMapping"]().translate(content)
    assert result == {
        "@context": CONTEXT,
        "type": "forge:Repository",
        "id": "https://github.com/progval/unicode_names2",
        "description": "char <-> Unicode character name (maintained fork of "
        "huonw/unicode_names)",
        "name": "progval/unicode_names2",
        "forge:forkedFrom": {
            "id": "https://github.com/huonw/unicode_names",
        },
    }


def test_github_issues():
    content = b"""
{
  "html_url": "https://github.com/SoftwareHeritage/swh-indexer",
  "has_issues": true
}
    """
    result = MAPPINGS["GitHubMapping"]().translate(content)
    assert result == {
        "@context": CONTEXT,
        "type": "forge:Repository",
        "id": "https://github.com/SoftwareHeritage/swh-indexer",
        "issueTracker": "https://github.com/SoftwareHeritage/swh-indexer/issues",
    }



# Copyright (C) 2023-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import calendar
import collections
import json
import secrets
import sys
from typing import Any, Dict, List, Optional
import uuid

import iso8601
from pybtex.database import Entry, Person
from pybtex.database.output.bibtex import Writer
from pybtex.plugin import register_plugin
from pyld.jsonld import JsonLdError
import rdflib

from swh.indexer.codemeta import CODEMETA_V3_CONTEXT_URL, compact, expand
from swh.indexer.metadata_dictionary.cff import CffMapping
from swh.indexer.namespaces import RDF, SCHEMA, SPDX_LICENSES
from swh.model.swhids import ObjectType, QualifiedSWHID

TMP_ROOT_URI_PREFIX = "https://www.softwareheritage.org/schema/2022/indexer/tmp-node/"
"""IRI used for `skolemization <https://www.w3.org/TR/rdf11-concepts/#section-skolemization>`_;
it is not used outside :func:`codemeta_to_bibtex`.
"""

MACRO_PREFIX = "macro" + secrets.token_urlsafe(16).replace("_", "")


class BibTeXCitationError(Exception):
    pass


class BibTeXWithMacroWriter(Writer):

    def quote(self, s):
        r"""
        >>> w = BibTeXWithMacroWriter()
        >>> print(w.quote(f'{MACRO_PREFIX}:jan'))
        jan
        """

        if s.startswith(f"{MACRO_PREFIX}:"):
            return s[len(MACRO_PREFIX) + 1 :]
        return super().quote(s)


register_plugin("pybtex.database.output", "bibtex_with_macro", BibTeXWithMacroWriter)


def codemeta_to_bibtex(
    doc: Dict[str, Any],
    swhid: Optional[QualifiedSWHID] = None,
    *,
    resolve_unknown_context_url: bool = False,
    force_codemeta_context: bool = False,
) -> str:
    """Generate citation in BibTeX format from a parsed ``codemeta.json`` file.

    Args:
        doc: parsed ``codemeta.json`` file
        swhid: optional SWHID to add as ``swhid`` field in BibTeX citation
        resolve_unknown_context_url: if const:`True` unknown JSON-LD context URL
            will be fetched using ``requests`` instead of raising an exception,
            :const:`False` by default as it can lead sending requests to arbitrary
            URLs so use with caution
        force_codemeta_context: if :const:`True`, the ``@context`` field in the
            JSON-LD document will be set to the CodeMeta v3.0 one, this can be used
            to ensure citation can be generated when strict JSON-LD parsing failed

    Returns:
        A BibTeX citation as a string.

    Raises:
        BibTeXCitationError: when citation could not be generated
    """

    if force_codemeta_context:
        doc["@context"] = CODEMETA_V3_CONTEXT_URL

    try:
        doc = compact(
            doc, forgefed=False, resolve_unknown_context_url=resolve_unknown_context_url
        )
    except JsonLdError as e:
        raise BibTeXCitationError(str(e.cause))

    identifiers = []

    if "id" in doc:
        identifiers.append(doc["id"])
    else:
        doc["id"] = f"_:{uuid.uuid4()}"

    id_: rdflib.term.Node
    if doc["id"].startswith("_:"):
        id_ = rdflib.term.BNode(doc["id"][2:])
    else:
        # using a base in case the id is not an absolute URI
        id_ = rdflib.term.URIRef(doc["id"], base=TMP_ROOT_URI_PREFIX)
        doc["id"] = str(id_)

    # workaround for https://github.com/codemeta/codemeta/pull/322
    if "identifier" in doc:
        if isinstance(doc["identifier"], list):
            for identifier in doc["identifier"]:
                if isinstance(identifier, str) and "/" not in identifier:
                    identifiers.append(identifier)
        elif isinstance(doc["identifier"], str) and "/" not in doc["identifier"]:
            identifiers.append(doc["identifier"])

    try:
        doc = expand(doc, resolve_unknown_context_url=resolve_unknown_context_url)
    except JsonLdError as e:
        raise BibTeXCitationError(str(e.cause))

    g = rdflib.Graph().parse(
        data=json.dumps(doc),
        format="json-ld",
        # replace invalid URIs with blank node ids, instead of discarding whole nodes:
        generalized_rdf=True,
    )
    persons: Dict[str, List[Person]] = collections.defaultdict(list)
    fields: Dict[str, Any] = {}

    def add_person(
        persons: List[Person],
        person_id: rdflib.term.Node,
        role_property: rdflib.term.URIRef,
    ) -> None:
        # If the node referenced by 'person_id' is actually a Role node, we need to look
        # deeper for the actual person node
        if (person_id, RDF.type, SCHEMA.Role) in g:
            for _, _, inner_person in g.triples((person_id, role_property, None)):
                add_person(persons, inner_person, role_property)

        person = Person()
        for _, _, name in g.triples((person_id, SCHEMA.name, None)):
            if (person_id, RDF.type, SCHEMA.Organization) in g:
                # prevent interpreting the name as "Firstname Lastname" and reformatting
                # it to "Lastname, Firstname"
                person.last_names.append(name)
            else:
                person = Person(name)

        for _, _, given_name in g.triples((person_id, SCHEMA.givenName, None)):
            person.first_names.append(given_name)

        for _, _, family_name in g.triples((person_id, SCHEMA.familyName, None)):
            person.last_names.append(family_name)

        if str(person) and person not in persons:
            persons.append(person)

    def add_affiliations(person: rdflib.term.Node) -> None:
        for _, _, organization in g.triples((person, SCHEMA.affiliation, None)):
            add_person(
                persons["organization"], organization, role_property=SCHEMA.affiliation
            )

    # abstract
    for _, _, description in g.triples((id_, SCHEMA.description, None)):
        fields["abstract"] = description
        break

    for _, _, author_or_author_list in g.triples((id_, SCHEMA.author, None)):
        # schema.org-style authors, which are single values
        add_person(
            persons["author"], author_or_author_list, role_property=SCHEMA.author
        )

        # codemeta-style authors, which are an ordered list
        if author_or_author_list == RDF.nil:
            # Workaround for https://github.com/RDFLib/rdflib/pull/2818
            continue
        for author in rdflib.collection.Collection(g, author_or_author_list):
            add_person(persons["author"], author, role_property=SCHEMA.author)
            add_affiliations(author)

    # date
    for _, _, date in g.triples((id_, SCHEMA.datePublished, None)):
        fields["date"] = date
        break
    else:
        for _, _, date in g.triples((id_, SCHEMA.dateCreated, None)):
            fields["date"] = date
            break
        else:
            for _, _, date in g.triples((id_, SCHEMA.dateModified, None)):
                fields["date"] = date
                break
    if "date" in fields:
        try:
            parsed_date = iso8601.parse_date(fields["date"])
            fields["year"] = str(parsed_date.year)
            fields["month"] = (
                f"{MACRO_PREFIX}:{calendar.month_abbr[parsed_date.month].lower()}"
            )
        except iso8601.ParseError:
            pass

    # identifier, doi, hal_id
    entry_key = None
    for _, _, identifier in g.triples((id_, SCHEMA.identifier, None)):
        identifiers.append(identifier)
    for identifier in identifiers:
        if entry_key is None and "/" not in identifier:
            # Avoid URLs
            entry_key = identifier
        if identifier.startswith("https://doi.org/"):
            fields["doi"] = identifier
        if identifier.startswith("hal-"):
            fields["hal_id"] = identifier

    # editor
    for _, _, editor in g.triples((id_, SCHEMA.editor, None)):
        add_person(persons["editor"], editor, role_property=SCHEMA.editor)
        add_affiliations(editor)

    # file
    for _, _, download_url in g.triples((id_, SCHEMA.downloadUrl, None)):
        fields["file"] = download_url
        break

    # license (represented by "Person" as it's the only way to make pybtex format
    # them as a list)
    for _, _, license in g.triples((id_, SCHEMA.license, None)):
        if license is None:
            continue
        license_ = str(license)
        if license_.startswith(str(SPDX_LICENSES)):
            license_ = license_[len(str(SPDX_LICENSES)) :]
            if license_.endswith(".html"):
                license_ = license_[:-5]
            persons["license"].append(Person(last=license_))

    # publisher
    for _, _, publisher in g.triples((id_, SCHEMA.publisher, None)):
        add_person(persons["publisher"], publisher, role_property=SCHEMA.publisher)
        add_affiliations(publisher)

    # repository
    for _, _, code_repository in g.triples((id_, SCHEMA.codeRepository, None)):
        fields["repository"] = code_repository
        break

    # title
    for _, _, name in g.triples((id_, SCHEMA.name, None)):
        fields["title"] = name
        break

    # url
    for _, _, name in g.triples((id_, SCHEMA.url, None)):
        fields["url"] = name
        break

    # version
    for _, _, version in g.triples((id_, SCHEMA.softwareVersion, None)):
        fields["version"] = version
        break
    else:
        for _, _, version in g.triples((id_, SCHEMA.version, None)):
            fields["version"] = version

    if not fields:
        raise BibTeXCitationError(
            "No BibTex fields could be extracted from citation metadata file "
            "(codemeta.json or citation.cff), please check its content is valid."
        )

    # entry_type
    if swhid:
        fields["swhid"] = str(swhid)
        if swhid.object_type == ObjectType.SNAPSHOT:
            entry_type = "software"
        elif swhid.object_type == ObjectType.CONTENT:
            entry_type = "codefragment"
        else:
            entry_type = "softwareversion"

        if entry_key is None:
            entry_key = f"swh-{swhid.object_type.value}-{swhid.object_id.hex()[:7]}"
            if swhid.lines:
                line_start, line_end = swhid.lines
                if line_start:
                    entry_key += f"-L{line_start}"
                if line_end:
                    entry_key += f"-L{line_end}"
    elif "version" in fields:
        entry_type = "softwareversion"
    else:
        entry_type = "software"

    entry = Entry(
        entry_type,
        persons=persons,
        fields=fields,
    )

    entry.key = entry_key or "REPLACEME"
    return entry.to_string(bib_format="bibtex_with_macro")


def cff_to_bibtex(content: str, swhid: Optional[QualifiedSWHID] = None) -> str:
    """Generate citation in BibTeX format from a raw ``citation.cff`` file.

    Args:
        content: raw content of a ``citation.cff`` file
        swhid: optional SWHID to add as ``swhid`` field in BibTeX citation

    Returns:
        A BibTeX citation as a string.

    Raises:
        BibTeXCitationError: when citation could not be generated
    """
    codemeta = CffMapping().translate(raw_content=content.encode("utf-8"))
    if codemeta is None:
        codemeta = {}
    return codemeta_to_bibtex(codemeta, swhid)


if __name__ == "__main__":
    for filename in sys.argv[1:]:
        if filename == "-":
            print(codemeta_to_bibtex(json.load(sys.stdin)))
        else:
            with open(filename) as f:
                print(codemeta_to_bibtex(json.load(f)))



# Copyright (C) 2017-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
import datetime
import functools
from typing import Any, Dict, List, Tuple, Union
import unittest

from hypothesis import strategies
import magic

from swh.core.api.classes import stream_results
from swh.indexer.storage import INDEXER_CFG_KEY
from swh.model.hashutil import HashDict, hash_to_bytes
from swh.model.model import (
    Content,
    Directory,
    DirectoryEntry,
    ObjectType,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Person,
    Release,
    Revision,
    RevisionType,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
    TimestampWithTimezone,
)
from swh.storage.utils import now

BASE_TEST_CONFIG: Dict[str, Dict[str, Any]] = {
    "storage": {"cls": "memory"},
    "objstorage": {"cls": "memory"},
    INDEXER_CFG_KEY: {"cls": "memory"},
}

ORIGIN_VISITS = [
    {"type": "git", "origin": "https://github.com/SoftwareHeritage/swh-storage"},
    {"type": "ftp", "origin": "rsync://ftp.gnu.org/gnu/3dldf"},
    {
        "type": "deposit",
        "origin": "https://forge.softwareheritage.org/source/jesuisgpl/",
    },
    {
        "type": "pypi",
        "origin": "https://old-pypi.example.org/project/limnoria/",
    },  # with rev head
    {"type": "pypi", "origin": "https://pypi.org/project/limnoria/"},  # with rel head
    {"type": "svn", "origin": "http://0-512-md.googlecode.com/svn/"},
    {"type": "git", "origin": "https://github.com/librariesio/yarn-parser"},
    {"type": "git", "origin": "https://github.com/librariesio/yarn-parser.git"},
    {"type": "git", "origin": "https://npm.example.org/yarn-parser"},
]

ORIGINS = [Origin(url=visit["origin"]) for visit in ORIGIN_VISITS]

OBJ_STORAGE_RAW_CONTENT: Dict[str, bytes] = {
    "text:some": b"this is some text",
    "text:another": b"another text",
    "text:yet": b"yet another text",
    "python:code": b"""
    import unittest
    import logging
    from swh.indexer.mimetype import MimetypeIndexer
    from swh.indexer.tests.test_utils import MockObjStorage

    class MockStorage():
        def content_mimetype_add(self, mimetypes):
            self.state = mimetypes

        def indexer_configuration_add(self, tools):
            return [{
                'id': 10,
            }]
    """,
    "c:struct": b"""
        #ifndef __AVL__
        #define __AVL__

        typedef struct _avl_tree avl_tree;

        typedef struct _data_t {
          int content;
        } data_t;
    """,
    "lisp:assertion": b"""
    (should 'pygments (recognize 'lisp 'easily))

    """,
    "json:test-metadata-package.json": b"""
    {
        "name": "test_metadata",
        "version": "0.0.1",
        "description": "Simple package.json test for indexer",
        "repository": {
          "type": "git",
          "url": "https://github.com/moranegg/metadata_test"
      }
    }
    """,
    "json:npm-package.json": b"""
    {
      "version": "5.0.3",
      "name": "npm",
      "description": "a package manager for JavaScript",
      "preferGlobal": true,
      "config": {
        "publishtest": false
      },
      "homepage": "https://docs.npmjs.com/",
      "author": "John Doe <john@doe.example.com> (http://blog.doe.example.com)",
      "repository": {
        "type": "git",
        "url": "https://github.com/npm/npm"
      },
      "bugs": {
        "url": "https://github.com/npm/npm/issues"
      },
      "dependencies": {
        "JSONStream": "~1.3.1",
        "abbrev": "~1.1.0",
        "ansi-regex": "~2.1.1",
        "ansicolors": "~0.3.2",
        "ansistyles": "~0.1.3"
      },
      "devDependencies": {
        "tacks": "~1.2.6",
        "tap": "~10.3.2"
      },
      "license": "Artistic-2.0"
    }

    """,
    "text:carriage-return": b"""
    """,
    "text:empty": b"",
    # was 626364 / b'bcd'
    "text:unimportant": b"unimportant content for bcd",
    # was 636465 / b'cde' now yarn-parser package.json
    "json:yarn-parser-package.json": b"""
    {
      "name": "yarn-parser",
      "version": "1.0.0",
      "description": "Tiny web service for parsing yarn.lock files",
      "main": "index.js",
      "scripts": {
        "start": "node index.js",
        "test": "mocha"
      },
      "engines": {
        "node": "9.8.0"
      },
      "repository": {
        "type": "git",
        "url": "git+https://github.com/librariesio/yarn-parser.git"
      },
      "author": "Gary Cooper",
      "license": "AGPL-3.0",
      "bugs": {
        "url": "https://github.com/librariesio/yarn-parser/issues"
      },
      "homepage": "https://github.com/librariesio/yarn-parser#readme",
      "dependencies": {
        "@yarnpkg/lockfile": "^1.0.0",
        "body-parser": "^1.15.2",
        "express": "^4.14.0"
      },
      "devDependencies": {
        "chai": "^4.1.2",
        "mocha": "^5.2.0",
        "request": "^2.87.0",
        "test": "^0.6.0"
      }
    }

""",
}

MAPPING_DESCRIPTION_CONTENT_SHA1GIT: Dict[str, bytes] = {}
MAPPING_DESCRIPTION_CONTENT_OBJID: Dict[str, HashDict] = {}
OBJ_STORAGE_DATA: List[Tuple[HashDict, bytes]] = []

for key_description, data in OBJ_STORAGE_RAW_CONTENT.items():
    content = Content.from_data(data)
    content_id = content.hashes()
    MAPPING_DESCRIPTION_CONTENT_SHA1GIT[key_description] = content.sha1_git
    MAPPING_DESCRIPTION_CONTENT_OBJID[key_description] = content_id
    OBJ_STORAGE_DATA.append((content_id, data))


RAW_CONTENT_METADATA: List[Tuple[bytes, Union[str, Tuple[str, ...]], str]] = [
    (
        "du français".encode(),
        "text/plain",
        "utf-8",
    ),
    (
        b"def __init__(self):",
        ("text/x-python", "text/x-script.python"),
        "us-ascii",
    ),
    (
        b"\xff\xfe\x00\x00\x00\x00\xff\xfe\xff\xff",
        "application/octet-stream",
        "" if magic.version() <= 544 else "utf-32le",
    ),
]

RAW_CONTENTS: List[Tuple[HashDict, bytes, Union[str, Tuple[str, ...]], str]] = []
RAW_CONTENT_OBJIDS: List[HashDict] = []

for index, raw_content_d in enumerate(RAW_CONTENT_METADATA):
    raw_content = raw_content_d[0]
    content = Content.from_data(raw_content)
    content_id = content.hashes()
    RAW_CONTENTS.append((content_id, *raw_content_d))
    RAW_CONTENT_OBJIDS.append(content_id)
    # and write it to objstorage data so it's flushed in the objstorage
    OBJ_STORAGE_DATA.append((content_id, raw_content))


SHA1_TO_LICENSES: Dict[bytes, List[str]] = {
    RAW_CONTENT_OBJIDS[0]["sha1"]: ["GPL"],
    RAW_CONTENT_OBJIDS[1]["sha1"]: ["AGPL"],
    RAW_CONTENT_OBJIDS[2]["sha1"]: [],
}


DIRECTORY = Directory(
    entries=(
        DirectoryEntry(
            name=b"index.js",
            type="file",
            target=MAPPING_DESCRIPTION_CONTENT_SHA1GIT["text:some"],
            perms=0o100644,
        ),
        DirectoryEntry(
            name=b"package.json",
            type="file",
            target=MAPPING_DESCRIPTION_CONTENT_SHA1GIT[
                "json:test-metadata-package.json"
            ],
            perms=0o100644,
        ),
        DirectoryEntry(
            name=b".github",
            type="dir",
            target=Directory(entries=()).id,
            perms=0o040000,
        ),
    ),
)

DIRECTORY2 = Directory(
    entries=(
        DirectoryEntry(
            name=b"package.json",
            type="file",
            target=MAPPING_DESCRIPTION_CONTENT_SHA1GIT["json:yarn-parser-package.json"],
            perms=0o100644,
        ),
    ),
)

_utc_plus_2 = datetime.timezone(datetime.timedelta(minutes=120))

REVISION = Revision(
    message=b"Improve search functionality",
    author=Person(
        name=b"Gary Cooper",
        fullname=b"Gary Cooper <gary.cooper@example.com>",
        email=b"gary.cooper@example.com",
    ),
    committer=Person(
        name=b"Gary Cooper",
        fullname=b"Gary Cooper <gary.cooper@example.com>",
        email=b"gary.cooper@example.com",
    ),
    committer_date=TimestampWithTimezone.from_datetime(
        datetime.datetime(2013, 10, 4, 12, 50, 49, tzinfo=_utc_plus_2)
    ),
    type=RevisionType.GIT,
    synthetic=False,
    date=TimestampWithTimezone.from_datetime(
        datetime.datetime(2017, 2, 20, 16, 14, 16, tzinfo=_utc_plus_2)
    ),
    directory=DIRECTORY2.id,
    parents=(),
)

REVISIONS = [REVISION]

RELEASE = Release(
    name=b"v0.0.0",
    message=None,
    author=Person(
        name=b"Gary Cooper",
        fullname=b"Gary Cooper <gary.cooper@example.com>",
        email=b"gary.cooper@example.com",
    ),
    synthetic=False,
    date=TimestampWithTimezone.from_datetime(
        datetime.datetime(2017, 2, 20, 16, 14, 16, tzinfo=_utc_plus_2)
    ),
    target_type=ObjectType.DIRECTORY,
    target=DIRECTORY2.id,
)

RELEASES = [RELEASE]

SNAPSHOTS = [
    # https://github.com/SoftwareHeritage/swh-storage
    Snapshot(
        branches={
            b"refs/heads/add-revision-origin-cache": SnapshotBranch(
                target=b'L[\xce\x1c\x88\x8eF\t\xf1"\x19\x1e\xfb\xc0s\xe7/\xe9l\x1e',
                target_type=SnapshotTargetType.REVISION,
            ),
            b"refs/head/master": SnapshotBranch(
                target=b"8K\x12\x00d\x03\xcc\xe4]bS\xe3\x8f{\xd7}\xac\xefrm",
                target_type=SnapshotTargetType.REVISION,
            ),
            b"HEAD": SnapshotBranch(
                target=b"refs/head/master", target_type=SnapshotTargetType.ALIAS
            ),
            b"refs/tags/v0.0.103": SnapshotBranch(
                target=b'\xb6"Im{\xfdLb\xb0\x94N\xea\x96m\x13x\x88+\x0f\xdd',
                target_type=SnapshotTargetType.RELEASE,
            ),
        },
    ),
    # rsync://ftp.gnu.org/gnu/3dldf
    Snapshot(
        branches={
            b"3DLDF-1.1.4.tar.gz": SnapshotBranch(
                target=b'dJ\xfb\x1c\x91\xf4\x82B%]6\xa2\x90|\xd3\xfc"G\x99\x11',
                target_type=SnapshotTargetType.REVISION,
            ),
            b"3DLDF-2.0.2.tar.gz": SnapshotBranch(
                target=b"\xb6\x0e\xe7\x9e9\xac\xaa\x19\x9e=\xd1\xc5\x00\\\xc6\xfc\xe0\xa6\xb4V",  # noqa
                target_type=SnapshotTargetType.REVISION,
            ),
            b"3DLDF-2.0.3-examples.tar.gz": SnapshotBranch(
                target=b"!H\x19\xc0\xee\x82-\x12F1\xbd\x97\xfe\xadZ\x80\x80\xc1\x83\xff",  # noqa
                target_type=SnapshotTargetType.REVISION,
            ),
            b"3DLDF-2.0.3.tar.gz": SnapshotBranch(
                target=b"\x8e\xa9\x8e/\xea}\x9feF\xf4\x9f\xfd\xee\xcc\x1a\xb4`\x8c\x8by",  # noqa
                target_type=SnapshotTargetType.REVISION,
            ),
            b"3DLDF-2.0.tar.gz": SnapshotBranch(
                target=b"F6*\xff(?\x19a\xef\xb6\xc2\x1fv$S\xe3G\xd3\xd1m",
                target_type=SnapshotTargetType.REVISION,
            ),
        },
    ),
    # https://forge.softwareheritage.org/source/jesuisgpl/",
    Snapshot(
        branches={
            b"master": SnapshotBranch(
                target=b"\xe7n\xa4\x9c\x9f\xfb\xb7\xf76\x11\x08{\xa6\xe9\x99\xb1\x9e]q\xeb",  # noqa
                target_type=SnapshotTargetType.REVISION,
            )
        },
    ),
    # https://old-pypi.example.org/project/limnoria/
    Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target=b"releases/2018.09.09", target_type=SnapshotTargetType.ALIAS
            ),
            b"releases/2018.09.01": SnapshotBranch(
                target=b"<\xee1(\xe8\x8d_\xc1\xc9\xa6rT\xf1\x1d\xbb\xdfF\xfdw\xcf",
                target_type=SnapshotTargetType.REVISION,
            ),
            b"releases/2018.09.09": SnapshotBranch(
                target=b"\x83\xb9\xb6\xc7\x05\xb1%\xd0\xfem\xd8kA\x10\x9d\xc5\xfa2\xf8t",  # noqa
                target_type=SnapshotTargetType.REVISION,
            ),
        },
    ),
    # https://pypi.org/project/limnoria/
    Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target=b"releases/2018.09.09", target_type=SnapshotTargetType.ALIAS
            ),
            b"releases/2018.09.01": SnapshotBranch(
                target=b"<\xee1(\xe8\x8d_\xc1\xc9\xa6rT\xf1\x1d\xbb\xdfF\xfdw\xcf",
                target_type=SnapshotTargetType.RELEASE,
            ),
            b"releases/2018.09.09": SnapshotBranch(
                target=b"\x83\xb9\xb6\xc7\x05\xb1%\xd0\xfem\xd8kA\x10\x9d\xc5\xfa2\xf8t",  # noqa
                target_type=SnapshotTargetType.RELEASE,
            ),
        },
    ),
    # http://0-512-md.googlecode.com/svn/
    Snapshot(
        branches={
            b"master": SnapshotBranch(
                target=b"\xe4?r\xe1,\x88\xab\xec\xe7\x9a\x87\xb8\xc9\xad#.\x1bw=\x18",
                target_type=SnapshotTargetType.REVISION,
            )
        },
    ),
    # https://github.com/librariesio/yarn-parser
    Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target=REVISION.id,
                target_type=SnapshotTargetType.REVISION,
            )
        },
    ),
    # https://github.com/librariesio/yarn-parser.git
    Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target=REVISION.id,
                target_type=SnapshotTargetType.REVISION,
            )
        },
    ),
    # https://npm.example.org/yarn-parser
    Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target=RELEASE.id,
                target_type=SnapshotTargetType.RELEASE,
            )
        },
    ),
]

assert len(SNAPSHOTS) == len(ORIGIN_VISITS)


YARN_PARSER_METADATA = {
    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
    "url": "https://github.com/librariesio/yarn-parser#readme",
    "codeRepository": "git+git+https://github.com/librariesio/yarn-parser.git",
    "author": [{"type": "Person", "name": "Gary Cooper"}],
    "license": "https://spdx.org/licenses/AGPL-3.0",
    "version": "1.0.0",
    "description": "Tiny web service for parsing yarn.lock files",
    "issueTracker": "https://github.com/librariesio/yarn-parser/issues",
    "name": "yarn-parser",
    "type": "SoftwareSourceCode",
}


json_dict_keys = strategies.one_of(
    strategies.characters(),
    strategies.just("type"),
    strategies.just("url"),
    strategies.just("name"),
    strategies.just("email"),
    strategies.just("@id"),
    strategies.just("@context"),
    strategies.just("repository"),
    strategies.just("license"),
    strategies.just("repositories"),
    strategies.just("licenses"),
)
"""Hypothesis strategy that generates strings, with an emphasis on those
that are often used as dictionary keys in metadata files."""


generic_json_document = strategies.recursive(
    strategies.none()
    | strategies.booleans()
    | strategies.floats()
    | strategies.characters(),
    lambda children: (
        strategies.lists(children, min_size=1)
        | strategies.dictionaries(json_dict_keys, children, min_size=1)
    ),
)
"""Hypothesis strategy that generates possible values for values of JSON
metadata files."""


def json_document_strategy(keys=None):
    """Generates an hypothesis strategy that generates metadata files
    for a JSON-based format that uses the given keys."""
    if keys is None:
        keys = strategies.characters()
    else:
        keys = strategies.one_of(map(strategies.just, keys))

    return strategies.dictionaries(keys, generic_json_document, min_size=1)


def _tree_to_xml(root, xmlns, data):
    def encode(s):
        "Skips unpaired surrogates generated by json_document_strategy"
        return s.encode("utf8", "replace")

    def to_xml(data, indent=b" "):
        if data is None:
            return b""
        elif isinstance(data, (bool, str, int, float)):
            return indent + encode(str(data))
        elif isinstance(data, list):
            return b"\n".join(to_xml(v, indent=indent) for v in data)
        elif isinstance(data, dict):
            lines = []
            for key, value in data.items():
                lines.append(indent + encode("<{}>".format(key)))
                lines.append(to_xml(value, indent=indent + b" "))
                lines.append(indent + encode("</{}>".format(key)))
            return b"\n".join(lines)
        else:
            raise TypeError(data)

    return b"\n".join(
        [
            '<{} xmlns="{}">'.format(root, xmlns).encode(),
            to_xml(data),
            "</{}>".format(root).encode(),
        ]
    )


class TreeToXmlTest(unittest.TestCase):
    def test_leaves(self):
        self.assertEqual(
            _tree_to_xml("root", "http://example.com", None),
            b'<root xmlns="http://example.com">\n\n</root>',
        )
        self.assertEqual(
            _tree_to_xml("root", "http://example.com", True),
            b'<root xmlns="http://example.com">\n True\n</root>',
        )
        self.assertEqual(
            _tree_to_xml("root", "http://example.com", "abc"),
            b'<root xmlns="http://example.com">\n abc\n</root>',
        )
        self.assertEqual(
            _tree_to_xml("root", "http://example.com", 42),
            b'<root xmlns="http://example.com">\n 42\n</root>',
        )
        self.assertEqual(
            _tree_to_xml("root", "http://example.com", 3.14),
            b'<root xmlns="http://example.com">\n 3.14\n</root>',
        )

    def test_dict(self):
        self.assertIn(
            _tree_to_xml("root", "http://example.com", {"foo": "bar", "baz": "qux"}),
            [
                b'<root xmlns="http://example.com">\n'
                b" <foo>\n  bar\n </foo>\n"
                b" <baz>\n  qux\n </baz>\n"
                b"</root>",
                b'<root xmlns="http://example.com">\n'
                b" <baz>\n  qux\n </baz>\n"
                b" <foo>\n  bar\n </foo>\n"
                b"</root>",
            ],
        )

    def test_list(self):
        self.assertEqual(
            _tree_to_xml(
                "root",
                "http://example.com",
                [
                    {"foo": "bar"},
                    {"foo": "baz"},
                ],
            ),
            b'<root xmlns="http://example.com">\n'
            b" <foo>\n  bar\n </foo>\n"
            b" <foo>\n  baz\n </foo>\n"
            b"</root>",
        )


def xml_document_strategy(keys, root, xmlns):
    """Generates an hypothesis strategy that generates metadata files
    for an XML format that uses the given keys."""

    return strategies.builds(
        functools.partial(_tree_to_xml, root, xmlns), json_document_strategy(keys)
    )


def filter_dict(d, keys):
    "return a copy of the dict with keys deleted"
    if not isinstance(keys, (list, tuple)):
        keys = (keys,)
    return dict((k, v) for (k, v) in d.items() if k not in keys)


def fill_obj_storage(obj_storage):
    """Add some content in an object storage."""
    for obj_id, content in OBJ_STORAGE_DATA:
        obj_storage.add(content, obj_id)


def fill_storage(storage):
    """Fill in storage with consistent test dataset."""
    storage.content_add([Content.from_data(data) for _, data in OBJ_STORAGE_DATA])
    storage.directory_add([DIRECTORY, DIRECTORY2])
    storage.revision_add(REVISIONS)
    storage.release_add(RELEASES)
    storage.snapshot_add(SNAPSHOTS)

    storage.origin_add(ORIGINS)
    for visit, snapshot in zip(ORIGIN_VISITS, SNAPSHOTS):
        assert snapshot.id is not None

        visit = storage.origin_visit_add(
            [OriginVisit(origin=visit["origin"], date=now(), type=visit["type"])]
        )[0]
        visit_status = OriginVisitStatus(
            origin=visit.origin,
            visit=visit.visit,
            date=now(),
            status="full",
            snapshot=snapshot.id,
        )
        storage.origin_visit_status_add([visit_status])


class CommonContentIndexerTest(metaclass=abc.ABCMeta):
    def get_indexer_results(self, ids):
        """Override this for indexers that don't have a mock storage."""
        return self.indexer.idx_storage.state

    def assert_results_ok(self, objids, expected_results=None):
        sha1s = [sha1["sha1"] for sha1 in objids]
        actual_results = list(self.get_indexer_results(sha1s))

        if expected_results is None:
            expected_results = self.expected_results

        # expected results may contain slightly duplicated results
        assert 0 < len(actual_results) <= len(expected_results)
        for result in actual_results:
            assert result in expected_results

    def test_index(self):
        """Known sha1 have their data indexed"""
        obj_ids = [self.id0, self.id1, self.id2]

        # when
        self.indexer.run(obj_ids)

        self.assert_results_ok(obj_ids)

        # 2nd pass
        self.indexer.run(obj_ids)

        self.assert_results_ok(obj_ids)

    def test_index_one_unknown_sha1(self):
        """Unknown sha1s are not indexed"""
        obj_ids = [
            self.id1,
            HashDict(
                sha1=bytes.fromhex("799a5ef812c53907562fe379d4b3851e69c7cb15")
            ),  # unknown
            HashDict(
                sha1=bytes.fromhex("800a5ef812c53907562fe379d4b3851e69c7cb15")
            ),  # unknown
        ]  # unknown

        # when
        self.indexer.run(obj_ids)

        # then
        expected_results = [
            res
            for res in self.expected_results
            if res.id in [sha1["sha1"] for sha1 in obj_ids]
        ]

        self.assert_results_ok(obj_ids, expected_results)


class CommonContentIndexerPartitionTest:
    """Allows to factorize tests on range indexer."""

    def setUp(self):
        self.contents = sorted(OBJ_STORAGE_DATA)

    def assert_results_ok(self, partition_id, nb_partitions, actual_results):
        expected_ids = [
            c.sha1
            for c in stream_results(
                self.indexer.storage.content_get_partition,
                partition_id=partition_id,
                nb_partitions=nb_partitions,
            )
        ]

        actual_results = list(actual_results)
        for indexed_data in actual_results:
            _id = indexed_data.id
            assert _id in expected_ids

            _tool_id = indexed_data.indexer_configuration_id
            assert _tool_id == self.indexer.tool["id"]

    def test__index_contents(self):
        """Indexing contents without existing data results in indexed data"""
        partition_id = 0
        nb_partitions = 4

        actual_results = list(
            self.indexer._index_contents(partition_id, nb_partitions, indexed={})
        )

        self.assert_results_ok(partition_id, nb_partitions, actual_results)

    def test__index_contents_with_indexed_data(self):
        """Indexing contents with existing data results in less indexed data"""
        partition_id = 3
        nb_partitions = 4

        # first pass
        actual_results = list(
            self.indexer._index_contents(partition_id, nb_partitions, indexed={}),
        )

        self.assert_results_ok(partition_id, nb_partitions, actual_results)

        indexed_ids = {res.id for res in actual_results}

        actual_results = list(
            self.indexer._index_contents(
                partition_id, nb_partitions, indexed=indexed_ids
            )
        )

        # already indexed, so nothing new
        assert actual_results == []

    def test_generate_content_get(self):
        """Optimal indexing should result in indexed data"""
        partition_id = 0
        nb_partitions = 1

        actual_results = self.indexer.run(
            partition_id, nb_partitions, skip_existing=False
        )

        assert actual_results["status"] == "eventful", actual_results

    def test_generate_content_get_no_result(self):
        """No result indexed returns False"""
        actual_results = self.indexer.run(1, 2**512, incremental=False)

        assert actual_results == {"status": "uneventful"}


def mock_compute_license(path):
    """path is the content identifier"""
    if isinstance(id, bytes):
        path = path.decode("utf-8")
    # path is something like /tmp/tmpXXX/<sha1> so we keep only the sha1 part
    id_ = path.split("/")[-1]
    return {"licenses": SHA1_TO_LICENSES.get(hash_to_bytes(id_), [])}



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os.path
from typing import Optional

from rdflib import BNode, Graph, Literal, URIRef

from swh.indexer.codemeta import _DATA_DIR, read_crosstable
from swh.indexer.namespaces import RDF, SCHEMA

from .base import JsonMapping, SingleFileIntrinsicMapping
from .utils import add_map

SPDX = URIRef("https://spdx.org/licenses/")


COMPOSER_TABLE_PATH = os.path.join(_DATA_DIR, "composer.csv")

with open(COMPOSER_TABLE_PATH) as fd:
    (CODEMETA_TERMS, COMPOSER_TABLE) = read_crosstable(fd)


class ComposerMapping(JsonMapping, SingleFileIntrinsicMapping):
    """Dedicated class for Packagist(composer.json) mapping and translation"""

    name = "composer"
    mapping = COMPOSER_TABLE["Composer"]
    filename = b"composer.json"
    string_fields = [
        "name",
        "description",
        "version",
        "keywords",
        "license",
        "author",
        "authors",
    ]
    uri_fields = ["homepage"]

    def normalize_license(self, s):
        if isinstance(s, str):
            return SPDX + s

    def _translate_author(self, graph: Graph, author) -> Optional[BNode]:
        if not isinstance(author, dict):
            return None
        node = BNode()
        graph.add((node, RDF.type, SCHEMA.Person))

        if isinstance(author.get("name"), str):
            graph.add((node, SCHEMA.name, Literal(author["name"])))
        if isinstance(author.get("email"), str):
            graph.add((node, SCHEMA.email, Literal(author["email"])))

        return node

    def translate_authors(self, graph: Graph, root: URIRef, authors) -> None:
        add_map(graph, root, SCHEMA.author, self._translate_author, authors)



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.indexer.metadata_dictionary import MAPPINGS


def test_compute_metadata_cff():
    """
    testing CITATION.cff translation
    """
    content = """# YAML 1.2
---
abstract: "Command line program to convert from Citation File \
Format to various other formats such as BibTeX, EndNote, RIS, \
schema.org, CodeMeta, and .zenodo.json."
authors:
  -
    affiliation: "Netherlands eScience Center"
    family-names: Fname
    given-names: Gname
  -
    affiliation: "Humboldt-Universität zu Berlin"
    family-names: Nachname
    given-names: Vorname
    orcid: https://orcid.org/0000-0003-0000-0007
cff-version: "1.0.3"
date-released: 2019-11-12
doi: 10.5281/zenodo.1162057
keywords:
  - "citation"
  - "bibliography"
  - "cff"
  - "CITATION.cff"
license: Apache-2.0
message: "If you use this software, please cite it using these metadata."
license: Apache-2.0
message: "If you use this software, please cite it using these metadata."
repository-code: "https://github.com/citation-file-format/cff-converter-python"
title: cffconvert
version: "1.4.0-alpha0"
    """.encode(
        "utf-8"
    )

    result = MAPPINGS["CffMapping"]().translate(content)
    assert set(result.pop("keywords")) == {
        "citation",
        "bibliography",
        "cff",
        "CITATION.cff",
    }
    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [
            {
                "type": "Person",
                "affiliation": {
                    "type": "Organization",
                    "name": "Netherlands eScience Center",
                },
                "familyName": "Fname",
                "givenName": "Gname",
            },
            {
                "id": "https://orcid.org/0000-0003-0000-0007",
                "type": "Person",
                "affiliation": {
                    "type": "Organization",
                    "name": "Humboldt-Universität zu Berlin",
                },
                "familyName": "Nachname",
                "givenName": "Vorname",
            },
        ],
        "codeRepository": (
            "https://github.com/citation-file-format/cff-converter-python"
        ),
        "datePublished": "2019-11-12",
        "description": """Command line program to convert from \
Citation File Format to various other formats such as BibTeX, EndNote, \
RIS, schema.org, CodeMeta, and .zenodo.json.""",
        "identifier": "https://doi.org/10.5281/zenodo.1162057",
        "license": "https://spdx.org/licenses/Apache-2.0",
        "version": "1.4.0-alpha0",
        "name": "cffconvert",
    }

    assert expected == result


def test_compute_metadata_cff_invalid_yaml():
    """
    test yaml translation for invalid yaml file
    """
    content = """cff-version: 1.0.3
message: To cite the SigMF specification, please include the following:
authors:
  - name: The GNU Radio Foundation, Inc.
    """.encode(
        "utf-8"
    )

    expected = None

    result = MAPPINGS["CffMapping"]().translate(content)
    assert expected == result


def test_compute_metadata_cff_empty():
    """
    test yaml translation for empty yaml file
    """
    content = """
    """.encode(
        "utf-8"
    )

    expected = None

    result = MAPPINGS["CffMapping"]().translate(content)
    assert expected == result


def test_compute_metadata_cff_list():
    """
    test yaml translation for empty yaml file
    """
    content = """
- Foo
- Bar
    """.encode(
        "utf-8"
    )

    expected = None

    result = MAPPINGS["CffMapping"]().translate(content)
    assert expected == result


def test_cff_empty_fields():
    """
    testing CITATION.cff translation
    """
    content = """# YAML 1.2
  authors:
  -
    affiliation: "Hogwarts"
    family-names:
    given-names: Harry
  -
    affiliation: "Ministry of Magic"
    family-names: Weasley
    orcid:
    given-names: Arthur


    """.encode(
        "utf-8"
    )

    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [
            {
                "type": "Person",
                "affiliation": {
                    "type": "Organization",
                    "name": "Hogwarts",
                },
                "givenName": "Harry",
            },
            {
                "type": "Person",
                "affiliation": {
                    "type": "Organization",
                    "name": "Ministry of Magic",
                },
                "familyName": "Weasley",
                "givenName": "Arthur",
            },
        ],
    }

    result = MAPPINGS["CffMapping"]().translate(content)
    assert expected == result


def test_cff_invalid_fields():
    """
    testing CITATION.cff translation
    """
    content = """# YAML 1.2
  authors:
  -
    affiliation: "Hogwarts"
    family-names:
    - Potter
    - James
    given-names: Harry

    """.encode(
        "utf-8"
    )

    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [
            {
                "type": "Person",
                "affiliation": {
                    "type": "Organization",
                    "name": "Hogwarts",
                },
                "givenName": "Harry",
            },
        ],
    }

    result = MAPPINGS["CffMapping"]().translate(content)
    assert expected == result


def test_cff_preferred_citation():
    content = """# YAML 1.2
cff-version: 1.2.0
title: scikit-learn
type: software
authors:
  - name: "The scikit-learn developers"
preferred-citation:
  type: article
  title: "Scikit-learn: Machine Learning in Python"
  authors:
  - family-names: "Pedregosa"
    given-names: "Fabian"
  - family-names: "Varoquaux"
    given-names: "Gaël"
  - family-names: "Gramfort"
    given-names: "Alexandre"
  - family-names: "Michel"
    given-names: "Vincent"
  - family-names: "Thirion"
    given-names: "Bertrand"
  - family-names: "Grisel"
    given-names: "Olivier"
  - family-names: "Blondel"
    given-names: "Mathieu"
  - family-names: "Prettenhofer"
    given-names: "Peter"
  - family-names: "Weiss"
    given-names: "Ron"
  - family-names: "Dubourg"
    given-names: "Vincent"
  - family-names: "Vanderplas"
    given-names: "Jake"
  - family-names: "Passos"
    given-names: "Alexandre"
  - family-names: "Cournapeau"
    given-names: "David"
  - family-names: "Brucher"
    given-names: "Matthieu"
  - family-names: "Perrot"
    given-names: "Matthieu"
  - family-names: "Duchesnay"
    given-names: "Édouard"
  journal: "Journal of Machine Learning Research"
  volume: 12
  start: 2825
  end: 2830
  year: 2011
  url: "https://jmlr.csail.mit.edu/papers/v12/pedregosa11a.html"
    """.encode(
        "utf-8"
    )

    expected = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "author": [
            {"type": "Person", "familyName": "Pedregosa", "givenName": "Fabian"},
            {"type": "Person", "familyName": "Varoquaux", "givenName": "Gaël"},
            {"type": "Person", "familyName": "Gramfort", "givenName": "Alexandre"},
            {"type": "Person", "familyName": "Michel", "givenName": "Vincent"},
            {"type": "Person", "familyName": "Thirion", "givenName": "Bertrand"},
            {"type": "Person", "familyName": "Grisel", "givenName": "Olivier"},
            {"type": "Person", "familyName": "Blondel", "givenName": "Mathieu"},
            {"type": "Person", "familyName": "Prettenhofer", "givenName": "Peter"},
            {"type": "Person", "familyName": "Weiss", "givenName": "Ron"},
            {"type": "Person", "familyName": "Dubourg", "givenName": "Vincent"},
            {"type": "Person", "familyName": "Vanderplas", "givenName": "Jake"},
            {"type": "Person", "familyName": "Passos", "givenName": "Alexandre"},
            {"type": "Person", "familyName": "Cournapeau", "givenName": "David"},
            {"type": "Person", "familyName": "Brucher", "givenName": "Matthieu"},
            {"type": "Person", "familyName": "Perrot", "givenName": "Matthieu"},
            {"type": "Person", "familyName": "Duchesnay", "givenName": "Édouard"},
        ],
        "name": "Scikit-learn: Machine Learning in Python",
        "url": "https://jmlr.csail.mit.edu/papers/v12/pedregosa11a.html",
    }

    result = MAPPINGS["CffMapping"]().translate(content)
    assert result == expected



# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

from hypothesis import HealthCheck, given, settings

from swh.indexer.metadata_dictionary import MAPPINGS

from ..utils import xml_document_strategy


def test_compute_metadata_maven():
    raw_content = b"""
    <project>
      <name>Maven Default Project</name>
      <modelVersion>4.0.0</modelVersion>
      <groupId>com.mycompany.app</groupId>
      <artifactId>my-app</artifactId>
      <version>1.2.3</version>
      <repositories>
        <repository>
          <id>central</id>
          <name>Maven Repository Switchboard</name>
          <layout>default</layout>
          <url>http://repo1.maven.org/maven2</url>
          <snapshots>
            <enabled>false</enabled>
          </snapshots>
        </repository>
      </repositories>
      <licenses>
        <license>
          <name>Apache License, Version 2.0</name>
          <url>https://www.apache.org/licenses/LICENSE-2.0.txt</url>
          <distribution>repo</distribution>
          <comments>A business-friendly OSS license</comments>
        </license>
      </licenses>
    </project>"""
    result = MAPPINGS["MavenMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "name": "Maven Default Project",
        "schema:identifier": "com.mycompany.app",
        "version": "1.2.3",
        "license": "https://www.apache.org/licenses/LICENSE-2.0.txt",
        "codeRepository": ("http://repo1.maven.org/maven2/com/mycompany/app/my-app"),
    }


def test_compute_metadata_maven_empty():
    raw_content = b"""
    <project>
    </project>"""
    result = MAPPINGS["MavenMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }


def test_compute_metadata_maven_almost_empty():
    raw_content = b"""
    <project>
      <foo/>
    </project>"""
    result = MAPPINGS["MavenMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
    }


def test_compute_metadata_maven_invalid_xml(caplog):
    expected_warning = (
        "swh.indexer.metadata_dictionary.maven.MavenMapping",
        logging.WARNING,
        "Error parsing XML from foo",
    )
    caplog.at_level(logging.WARNING, logger="swh.indexer.metadata_dictionary")

    raw_content = b"""
    <project>"""
    caplog.clear()
    result = MAPPINGS["MavenMapping"]("foo").translate(raw_content)
    assert caplog.record_tuples == [expected_warning], result
    assert result is None

    raw_content = b"""
    """
    caplog.clear()
    result = MAPPINGS["MavenMapping"]("foo").translate(raw_content)
    assert caplog.record_tuples == [expected_warning], result
    assert result is None


def test_compute_metadata_maven_unknown_encoding(caplog):
    expected_warning = (
        "swh.indexer.metadata_dictionary.maven.MavenMapping",
        logging.WARNING,
        "Error detecting XML encoding from foo",
    )
    caplog.at_level(logging.WARNING, logger="swh.indexer.metadata_dictionary")

    raw_content = b"""<?xml version="1.0" encoding="foo"?>
    <project>
    </project>"""
    caplog.clear()
    result = MAPPINGS["MavenMapping"]("foo").translate(raw_content)
    assert caplog.record_tuples == [expected_warning], result
    assert result is None

    raw_content = b"""<?xml version="1.0" encoding="UTF-7"?>
    <project>
    </project>"""
    caplog.clear()
    result = MAPPINGS["MavenMapping"]("foo").translate(raw_content)
    assert caplog.record_tuples == [expected_warning], result
    assert result is None


def test_compute_metadata_maven_invalid_encoding(caplog):
    expected_warning = [
        # libexpat1 <= 2.2.10-2+deb11u1
        [
            (
                "swh.indexer.metadata_dictionary.maven.MavenMapping",
                logging.WARNING,
                "Error unidecoding XML from foo",
            )
        ],
        # libexpat1 >= 2.2.10-2+deb11u2
        [
            (
                "swh.indexer.metadata_dictionary.maven.MavenMapping",
                logging.WARNING,
                "Error parsing XML from foo",
            )
        ],
    ]
    caplog.at_level(logging.WARNING, logger="swh.indexer.metadata_dictionary")

    raw_content = b"""<?xml version="1.0" encoding="UTF-8"?>
    <foo\xe5ct>
    </foo>"""
    caplog.clear()
    result = MAPPINGS["MavenMapping"]("foo").translate(raw_content)
    assert caplog.record_tuples in expected_warning, result
    assert result is None


def test_compute_metadata_maven_minimal():
    raw_content = b"""
    <project>
      <name>Maven Default Project</name>
      <modelVersion>4.0.0</modelVersion>
      <groupId>com.mycompany.app</groupId>
      <artifactId>my-app</artifactId>
      <version>1.2.3</version>
    </project>"""
    result = MAPPINGS["MavenMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "name": "Maven Default Project",
        "schema:identifier": "com.mycompany.app",
        "version": "1.2.3",
        "codeRepository": (
            "https://repo.maven.apache.org/maven2/com/mycompany/app/my-app"
        ),
    }


def test_compute_metadata_maven_empty_nodes():
    raw_content = b"""
    <project>
      <name>Maven Default Project</name>
      <modelVersion>4.0.0</modelVersion>
      <groupId>com.mycompany.app</groupId>
      <artifactId>my-app</artifactId>
      <version>1.2.3</version>
      <repositories>
      </repositories>
    </project>"""
    result = MAPPINGS["MavenMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "name": "Maven Default Project",
        "schema:identifier": "com.mycompany.app",
        "version": "1.2.3",
        "codeRepository": (
            "https://repo.maven.apache.org/maven2/com/mycompany/app/my-app"
        ),
    }

    raw_content = b"""
    <project>
      <name>Maven Default Project</name>
      <modelVersion>4.0.0</modelVersion>
      <groupId>com.mycompany.app</groupId>
      <artifactId>my-app</artifactId>
      <version></version>
    </project>"""
    result = MAPPINGS["MavenMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "name": "Maven Default Project",
        "schema:identifier": "com.mycompany.app",
        "codeRepository": (
            "https://repo.maven.apache.org/maven2/com/mycompany/app/my-app"
        ),
    }

    raw_content = b"""
    <project>
      <name></name>
      <modelVersion>4.0.0</modelVersion>
      <groupId>com.mycompany.app</groupId>
      <artifactId>my-app</artifactId>
      <version>1.2.3</version>
    </project>"""
    result = MAPPINGS["MavenMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "schema:identifier": "com.mycompany.app",
        "version": "1.2.3",
        "codeRepository": (
            "https://repo.maven.apache.org/maven2/com/mycompany/app/my-app"
        ),
    }

    raw_content = b"""
    <project>
      <name>Maven Default Project</name>
      <modelVersion>4.0.0</modelVersion>
      <groupId>com.mycompany.app</groupId>
      <artifactId>my-app</artifactId>
      <version>1.2.3</version>
      <licenses>
      </licenses>
    </project>"""
    result = MAPPINGS["MavenMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "name": "Maven Default Project",
        "schema:identifier": "com.mycompany.app",
        "version": "1.2.3",
        "codeRepository": (
            "https://repo.maven.apache.org/maven2/com/mycompany/app/my-app"
        ),
    }

    raw_content = b"""
    <project>
      <groupId></groupId>
      <version>1.2.3</version>
    </project>"""
    result = MAPPINGS["MavenMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "version": "1.2.3",
    }


def test_compute_metadata_maven_invalid_licenses():
    raw_content = b"""
    <project>
      <name>Maven Default Project</name>
      <modelVersion>4.0.0</modelVersion>
      <groupId>com.mycompany.app</groupId>
      <artifactId>my-app</artifactId>
      <version>1.2.3</version>
      <licenses>
        foo
      </licenses>
    </project>"""
    result = MAPPINGS["MavenMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "name": "Maven Default Project",
        "schema:identifier": "com.mycompany.app",
        "version": "1.2.3",
        "codeRepository": (
            "https://repo.maven.apache.org/maven2/com/mycompany/app/my-app"
        ),
    }


def test_compute_metadata_maven_multiple():
    """Tests when there are multiple code repos and licenses."""
    raw_content = b"""
    <project>
      <name>Maven Default Project</name>
      <modelVersion>4.0.0</modelVersion>
      <groupId>com.mycompany.app</groupId>
      <artifactId>my-app</artifactId>
      <version>1.2.3</version>
      <repositories>
        <repository>
          <id>central</id>
          <name>Maven Repository Switchboard</name>
          <layout>default</layout>
          <url>http://repo1.maven.org/maven2</url>
          <snapshots>
            <enabled>false</enabled>
          </snapshots>
        </repository>
        <repository>
          <id>example</id>
          <name>Example Maven Repo</name>
          <layout>default</layout>
          <url>http://example.org/maven2</url>
        </repository>
      </repositories>
      <licenses>
        <license>
          <name>Apache License, Version 2.0</name>
          <url>https://www.apache.org/licenses/LICENSE-2.0.txt</url>
          <distribution>repo</distribution>
          <comments>A business-friendly OSS license</comments>
        </license>
        <license>
          <name>MIT license</name>
          <url>https://opensource.org/licenses/MIT</url>
        </license>
      </licenses>
    </project>"""
    result = MAPPINGS["MavenMapping"]().translate(raw_content)
    assert set(result.pop("license")) == {
        "https://www.apache.org/licenses/LICENSE-2.0.txt",
        "https://opensource.org/licenses/MIT",
    }, result
    assert set(result.pop("codeRepository")) == {
        "http://repo1.maven.org/maven2/com/mycompany/app/my-app",
        "http://example.org/maven2/com/mycompany/app/my-app",
    }, result
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "name": "Maven Default Project",
        "schema:identifier": "com.mycompany.app",
        "version": "1.2.3",
    }


def test_compute_metadata_maven_invalid_repository():
    raw_content = b"""
    <project>
      <name>Maven Default Project</name>
      <modelVersion>4.0.0</modelVersion>
      <groupId>com.mycompany.app</groupId>
      <artifactId>my-app</artifactId>
      <version>1.2.3</version>
      <repositories>
        <repository>
          <id>tcc-transaction-internal-releases</id>
          <name>internal repository for released artifacts</name>
          <url>${repo.internal.releases.url}</url>
          <snapshots>
              <enabled>false</enabled>
          </snapshots>
          <releases>
              <enabled>true</enabled>
          </releases>
        </repository>
      </repositories>
      <licenses>
        <license>
          <name>Apache License, Version 2.0</name>
          <url>https://www.apache.org/licenses/LICENSE-2.0.txt</url>
          <distribution>repo</distribution>
          <comments>A business-friendly OSS license</comments>
        </license>
      </licenses>
    </project>"""
    result = MAPPINGS["MavenMapping"]().translate(raw_content)
    assert result == {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "type": "SoftwareSourceCode",
        "name": "Maven Default Project",
        "schema:identifier": "com.mycompany.app",
        "version": "1.2.3",
        "license": "https://www.apache.org/licenses/LICENSE-2.0.txt",
    }


@settings(suppress_health_check=[HealthCheck.too_slow])
@given(
    xml_document_strategy(
        keys=list(MAPPINGS["MavenMapping"].mapping),  # type: ignore
        root="project",
        xmlns="http://maven.apache.org/POM/4.0.0",
    )
)
def test_maven_adversarial(doc):
    MAPPINGS["MavenMapping"]().translate(doc)



# Copyright (C) 2020-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Classes used internally by the in-memory idx-storage, and will be
used for the interface of the idx-storage in the near future."""

from __future__ import annotations

import json
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar

import attr
from typing_extensions import Final

from swh.model.model import Sha1Git

TSelf = TypeVar("TSelf")


@attr.s
class BaseRow:
    UNIQUE_KEY_FIELDS: Tuple = ("id",)

    id = attr.ib(type=Any)
    indexer_configuration_id = attr.ib(type=Optional[int], default=None, kw_only=True)
    tool = attr.ib(type=Optional[Dict], default=None, kw_only=True)

    def __attrs_post_init__(self):
        if self.indexer_configuration_id is None and self.tool is None:
            raise TypeError("Either indexer_configuration_id or tool must be not None.")
        if self.indexer_configuration_id is not None and self.tool is not None:
            raise TypeError(
                "indexer_configuration_id and tool are mutually exclusive; "
                "only one may be not None."
            )

    def anonymize(self: TSelf) -> Optional[TSelf]:
        # Needed to implement swh.journal.writer.ValueProtocol
        return None

    def to_dict(self) -> Dict[str, Any]:
        """Wrapper of `attr.asdict` that can be overridden by subclasses
        that have special handling of some of the fields."""
        d = attr.asdict(self, recurse=False)

        if d["indexer_configuration_id"] is None:
            del d["indexer_configuration_id"]
        if d["tool"] is None:
            del d["tool"]

        return d

    @classmethod
    def from_dict(cls: Type[TSelf], d) -> TSelf:
        return cls(**d)

    def unique_key(self) -> Dict:
        if not self.tool:
            raise ValueError(
                f"Cannot compute unique_key of {self.__class__.__name__} with no tool "
                f"dictionary (indexer_configuration_id was given instead)"
            )

        tool_dict = {
            "tool_name": self.tool["name"],
            "tool_version": self.tool["version"],
            "tool_configuration": json.dumps(
                self.tool["configuration"], sort_keys=True
            ),
        }

        return {
            **{key: getattr(self, key) for key in self.UNIQUE_KEY_FIELDS},
            **tool_dict,
        }


@attr.s
class ContentMimetypeRow(BaseRow):
    object_type: Final = "content_mimetype"

    id = attr.ib(type=Sha1Git)
    mimetype = attr.ib(type=str)
    encoding = attr.ib(type=str)


@attr.s
class ContentLicenseRow(BaseRow):
    object_type: Final = "content_fossology_license"
    UNIQUE_KEY_FIELDS = ("id", "license")

    id = attr.ib(type=Sha1Git)
    license = attr.ib(type=str)


@attr.s
class ContentMetadataRow(BaseRow):
    object_type: Final = "content_metadata"

    id = attr.ib(type=Sha1Git)
    metadata = attr.ib(type=Dict[str, Any])


@attr.s
class DirectoryIntrinsicMetadataRow(BaseRow):
    object_type: Final = "directory_intrinsic_metadata"

    id = attr.ib(type=Sha1Git)
    metadata = attr.ib(type=Dict[str, Any])
    mappings = attr.ib(type=List[str])


@attr.s
class OriginIntrinsicMetadataRow(BaseRow):
    object_type: Final = "origin_intrinsic_metadata"

    id = attr.ib(type=str)
    metadata = attr.ib(type=Dict[str, Any])
    from_directory = attr.ib(type=Sha1Git)
    mappings = attr.ib(type=List[str])


@attr.s
class OriginExtrinsicMetadataRow(BaseRow):
    object_type: Final = "origin_extrinsic_metadata"

    id = attr.ib(type=str)
    """origin URL"""
    metadata = attr.ib(type=Dict[str, Any])
    from_remd_id = attr.ib(type=Sha1Git)
    """id of the RawExtrinsicMetadata object used as source for indexed metadata"""
    mappings = attr.ib(type=List[str])



# Copyright (C) 2020-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict, Iterable, List, Optional
from unittest.mock import Mock

import pytest
import sentry_sdk

from swh.indexer.indexer import ContentIndexer, DirectoryIndexer, OriginIndexer
from swh.indexer.storage import Sha1
from swh.model.hashutil import HashDict

from .utils import BASE_TEST_CONFIG, DIRECTORY2


class _TestException(Exception):
    pass


class CrashingIndexerMixin:
    USE_TOOLS = False

    def index(
        self, id: Any, data: Optional[Any] = None, **kwargs
    ) -> List[Dict[str, Any]]:
        raise _TestException()

    def persist_index_computations(self, results) -> Dict[str, int]:
        return {}

    def indexed_contents_in_partition(
        self, partition_id: int, nb_partitions: int
    ) -> Iterable[Sha1]:
        raise _TestException()


class CrashingContentIndexer(CrashingIndexerMixin, ContentIndexer):
    pass


class CrashingDirectoryIndexer(CrashingIndexerMixin, DirectoryIndexer):
    pass


class CrashingOriginIndexer(CrashingIndexerMixin, OriginIndexer):
    pass


def check_sentry(sentry_events, tags):
    assert len(sentry_events) == 1
    sentry_event = sentry_events.pop()
    assert sentry_event.get("tags") == tags
    assert "'_TestException'" in str(sentry_event)


def test_content_indexer_catch_exceptions(sentry_events):
    indexer = CrashingContentIndexer(config=BASE_TEST_CONFIG)
    indexer.objstorage = Mock()
    indexer.objstorage.get.return_value = b"content"
    indexer.objstorage.get_batch.return_value = [b"content"]

    sha1 = b"\x12" * 20

    # As task, catching exceptions
    assert indexer.run([HashDict(sha1=sha1)]) == ({"status": "failed"}, [])
    check_sentry(sentry_events, {"swh-indexer-content-sha1": sha1.hex()})

    # As journal client, catching exceptions
    assert indexer.process_journal_objects({"content": [{"sha1": sha1}]}) == {
        "status": "failed"
    }
    check_sentry(sentry_events, {"swh-indexer-content-sha1": sha1.hex()})

    indexer.catch_exceptions = False

    # As task, not catching exceptions
    with pytest.raises(_TestException):
        indexer.run([HashDict(sha1=sha1)])
    assert sentry_events == []

    # As journal client, not catching exceptions
    with pytest.raises(_TestException):
        indexer.process_journal_objects({"content": [{"sha1": sha1}]})
    assert sentry_events == []

    # As journal client, check the frontend will be able to get the tag when reporting
    try:
        indexer.process_journal_objects({"content": [{"sha1": sha1}]})
    except Exception:
        sentry_sdk.capture_exception()
    else:
        assert False
    check_sentry(sentry_events, {"swh-indexer-content-sha1": sha1.hex()})


def test_directory_indexer_catch_exceptions(sentry_events):
    indexer = CrashingDirectoryIndexer(config=BASE_TEST_CONFIG)
    indexer.storage = Mock()
    indexer.storage.directory_get.return_value = [DIRECTORY2]

    sha1 = DIRECTORY2.id
    swhid = str(DIRECTORY2.swhid())

    # As task, catching exceptions
    assert indexer.run([sha1]) == ({"status": "failed"}, [])
    check_sentry(sentry_events, {"swh-indexer-directory-swhid": swhid})

    # As journal client, catching exceptions
    assert indexer.process_journal_objects({"directory": [DIRECTORY2.to_dict()]}) == {
        "status": "failed"
    }
    check_sentry(sentry_events, {"swh-indexer-directory-swhid": swhid})

    indexer.catch_exceptions = False

    # As task, not catching exceptions
    with pytest.raises(_TestException):
        indexer.run([b"foo"])
    assert sentry_events == []

    # As journal client, not catching exceptions
    with pytest.raises(_TestException):
        indexer.process_journal_objects({"directory": [DIRECTORY2.to_dict()]})
    assert sentry_events == []

    # As journal client, check the frontend will be able to get the tag when reporting
    try:
        indexer.process_journal_objects({"directory": [DIRECTORY2.to_dict()]})
    except Exception:
        sentry_sdk.capture_exception()
    else:
        assert False
    check_sentry(sentry_events, {"swh-indexer-directory-swhid": swhid})


def test_origin_indexer_catch_exceptions(sentry_events):
    indexer = CrashingOriginIndexer(config=BASE_TEST_CONFIG)

    origin_url = "http://example.org"

    # As task, catching exceptions
    assert indexer.run([origin_url]) == ({"status": "failed"}, [])
    check_sentry(sentry_events, {"swh-indexer-origin-url": origin_url})

    # As journal client, catching exceptions
    assert indexer.process_journal_objects({"origin": [{"url": origin_url}]}) == {
        "status": "failed"
    }
    check_sentry(sentry_events, {"swh-indexer-origin-url": origin_url})

    indexer.catch_exceptions = False

    # As task, not catching exceptions
    with pytest.raises(_TestException):
        indexer.run([origin_url])
    assert sentry_events == []

    # As journal client, not catching exceptions
    with pytest.raises(_TestException):
        indexer.process_journal_objects({"origin": [{"url": origin_url}]})
    assert sentry_events == []

    # As journal client, check the frontend will be able to get the tag when reporting
    try:
        indexer.process_journal_objects({"origin": [{"url": origin_url}]})
    except Exception:
        sentry_sdk.capture_exception()
    else:
        assert False
    check_sentry(sentry_events, {"swh-indexer-origin-url": origin_url})



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from functools import wraps
import logging

from swh.core.statsd import statsd

OPERATIONS_METRIC = "swh_indexer_storage_operations_total"
OPERATIONS_UNIT_METRIC = "swh_indexer_storage_operations_{unit}_total"
DURATION_METRIC = "swh_indexer_storage_request_duration_seconds"


def timed(f):
    """Time that function!"""

    @wraps(f)
    def d(*a, **kw):
        with statsd.timed(DURATION_METRIC, tags={"endpoint": f.__name__}):
            return f(*a, **kw)

    return d


def send_metric(metric, count, method_name):
    """Send statsd metric with count for method `method_name`

    If count is 0, the metric is discarded.  If the metric is not
    parseable, the metric is discarded with a log message.

    Args:
        metric (str): Metric's name (e.g content:add, content:add:bytes)
        count (int): Associated value for the metric
        method_name (str): Method's name

    Returns:
        Bool to explicit if metric has been set or not
    """
    if count == 0:
        return False

    metric_type = metric.split(":")
    _length = len(metric_type)
    if _length == 2:
        object_type, operation = metric_type
        metric_name = OPERATIONS_METRIC
    elif _length == 3:
        object_type, operation, unit = metric_type
        metric_name = OPERATIONS_UNIT_METRIC.format(unit=unit)
    else:
        logging.warning("Skipping unknown metric {%s: %s}" % (metric, count))
        return False

    statsd.increment(
        metric_name,
        count,
        tags={
            "endpoint": method_name,
            "object_type": object_type,
            "operation": operation,
        },
    )
    return True


def process_metrics(f):
    """Increment object counters for the decorated function."""

    @wraps(f)
    def d(*a, **kw):
        r = f(*a, **kw)
        for metric, count in r.items():
            send_metric(metric=metric, count=count, method_name=f.__name__)

        return r

    return d



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import patch

from swh.indexer.storage.metrics import (
    OPERATIONS_METRIC,
    OPERATIONS_UNIT_METRIC,
    send_metric,
)


def test_send_metric_unknown_unit() -> None:
    r = send_metric("content", count=10, method_name="content_add")
    assert r is False
    r = send_metric("sthg:add:bytes:extra", count=10, method_name="sthg_add")
    assert r is False


def test_send_metric_no_value() -> None:
    r = send_metric("content_mimetype:add", count=0, method_name="content_mimetype_add")
    assert r is False


@patch("swh.indexer.storage.metrics.statsd.increment")
def test_send_metric_no_unit(mock_statsd) -> None:
    r = send_metric(
        "content_mimetype:add", count=10, method_name="content_mimetype_add"
    )

    mock_statsd.assert_called_with(
        OPERATIONS_METRIC,
        10,
        tags={
            "endpoint": "content_mimetype_add",
            "object_type": "content_mimetype",
            "operation": "add",
        },
    )

    assert r


@patch("swh.indexer.storage.metrics.statsd.increment")
def test_send_metric_unit(mock_statsd) -> None:
    unit_ = "bytes"
    r = send_metric("c:add:%s" % unit_, count=100, method_name="c_add")

    expected_metric = OPERATIONS_UNIT_METRIC.format(unit=unit_)
    mock_statsd.assert_called_with(
        expected_metric,
        100,
        tags={
            "endpoint": "c_add",
            "object_type": "c",
            "operation": "add",
        },
    )

    assert r



# Copyright (C) 2015-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


def db_to_mimetype(mimetype):
    """Convert a mimetype entry into a ready mimetype output."""
    return {
        "id": mimetype["id"],
        "encoding": mimetype["encoding"],
        "mimetype": mimetype["mimetype"],
        "tool": {
            "id": mimetype["tool_id"],
            "name": mimetype["tool_name"],
            "version": mimetype["tool_version"],
            "configuration": mimetype["tool_configuration"],
        },
    }


def db_to_metadata(metadata):
    """Convert a metadata entry into a ready metadata output."""
    metadata["tool"] = {
        "id": metadata["tool_id"],
        "name": metadata["tool_name"],
        "version": metadata["tool_version"],
        "configuration": metadata["tool_configuration"],
    }
    del metadata["tool_id"], metadata["tool_configuration"]
    del metadata["tool_version"], metadata["tool_name"]
    return metadata


def db_to_fossology_license(license):
    return {
        "id": license["id"],
        "license": license["license"],
        "tool": {
            "id": license["tool_id"],
            "name": license["tool_name"],
            "version": license["tool_version"],
            "configuration": license["tool_configuration"],
        },
    }



# Copyright (C) 2017-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict
import unittest

import pytest

from swh.indexer.mimetype import MimetypeIndexer, compute_mimetype_encoding
from swh.indexer.storage.model import ContentMimetypeRow
from swh.indexer.tests.utils import (
    BASE_TEST_CONFIG,
    RAW_CONTENT_OBJIDS,
    RAW_CONTENTS,
    CommonContentIndexerTest,
    fill_obj_storage,
    fill_storage,
    filter_dict,
)


@pytest.mark.parametrize(
    "content_id,raw_text,mimetypes,encoding",
    RAW_CONTENTS,
)
def test_compute_mimetype_encoding(content_id, raw_text, mimetypes, encoding):
    """Compute mimetype encoding should return results"""
    actual_result = compute_mimetype_encoding(raw_text)

    # Older libmagic versions (e.g. buster: 1:5.35-4+deb10u2, bullseye: 1:5.39-3)
    # returns different results. This allows to deal with such a case when executing
    # tests on different environments machines (e.g. ci tox, ci debian, dev machine,
    # ...)
    all_mimetypes = mimetypes if isinstance(mimetypes, tuple) else [mimetypes]

    assert actual_result in [
        {"mimetype": mimetype, "encoding": encoding} for mimetype in all_mimetypes
    ]


CONFIG: Dict[str, Any] = {
    **BASE_TEST_CONFIG,
    "tools": {
        "name": "file",
        "version": "1:5.30-1+deb9u1",
        "configuration": {"type": "library", "debian-package": "python3-magic"},
    },
}


class TestMimetypeIndexer(CommonContentIndexerTest, unittest.TestCase):
    """Mimetype indexer test scenarios:

    - Known sha1s in the input list have their data indexed
    - Unknown sha1 in the input list are not indexed

    """

    def get_indexer_results(self, ids):
        yield from self.idx_storage.content_mimetype_get(ids)

    def setUp(self):
        self.indexer = MimetypeIndexer(config=CONFIG)
        self.indexer.catch_exceptions = False
        self.idx_storage = self.indexer.idx_storage
        fill_storage(self.indexer.storage)
        fill_obj_storage(self.indexer.objstorage)

        self.id0, self.id1, self.id2 = RAW_CONTENT_OBJIDS

        tool = {k.replace("tool_", ""): v for (k, v) in self.indexer.tool.items()}

        results = []
        for raw_content_id, raw_content, mimetypes, encoding in RAW_CONTENTS:
            # Older libmagic versions (e.g. buster: 1:5.35-4+deb10u2, bullseye:
            # 1:5.39-3) returns different results. This allows to deal with such a case
            # when executing tests on different environments machines (e.g. ci tox, ci
            # debian, dev machine, ...)
            all_mimetypes = mimetypes if isinstance(mimetypes, tuple) else [mimetypes]

            results.extend(
                [
                    ContentMimetypeRow(
                        id=raw_content_id["sha1"],
                        tool=tool,
                        mimetype=mimetype,
                        encoding=encoding,
                    )
                    for mimetype in all_mimetypes
                ]
            )

        self.expected_results = results


RANGE_CONFIG = dict(list(CONFIG.items()) + [("write_batch_size", 100)])


def test_mimetype_w_no_tool():
    with pytest.raises(ValueError):
        MimetypeIndexer(config=filter_dict(CONFIG, "tools"))



# Copyright (C) 2018-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import re

from rdflib import RDF, BNode, Graph, Literal, URIRef

from swh.indexer.codemeta import CROSSWALK_TABLE
from swh.indexer.namespaces import SCHEMA

from .base import JsonMapping, SingleFileIntrinsicMapping
from .utils import add_list, add_url_if_valid, prettyprint_graph  # noqa

SPDX = URIRef("https://spdx.org/licenses/")


class NpmMapping(JsonMapping, SingleFileIntrinsicMapping):
    """
    dedicated class for NPM (package.json) mapping and translation
    """

    name = "npm"
    mapping = CROSSWALK_TABLE["NodeJS"]
    filename = b"package.json"
    string_fields = ["name", "version", "description", "email"]
    uri_fields = ["homepage"]

    _schema_shortcuts = {
        "github": "git+https://github.com/%s.git",
        "gist": "git+https://gist.github.com/%s.git",
        "gitlab": "git+https://gitlab.com/%s.git",
        # Bitbucket supports both hg and git, and the shortcut does not
        # tell which one to use.
        # 'bitbucket': 'https://bitbucket.org/',
    }

    def normalize_repository(self, d):
        """https://docs.npmjs.com/cli/v11/configuring-npm/package-json#repository

        >>> NpmMapping().normalize_repository({
        ...     'type': 'git',
        ...     'url': 'https://example.org/foo.git'
        ... })
        rdflib.term.URIRef('git+https://example.org/foo.git')
        >>> NpmMapping().normalize_repository(
        ...     'gitlab:foo/bar')
        rdflib.term.URIRef('git+https://gitlab.com/foo/bar.git')
        >>> NpmMapping().normalize_repository(
        ...     'foo/bar')
        rdflib.term.URIRef('git+https://github.com/foo/bar.git')
        """
        if (
            isinstance(d, dict)
            and isinstance(d.get("type"), str)
            and isinstance(d.get("url"), str)
        ):
            url = "{type}+{url}".format(**d)
        elif isinstance(d, str):
            if "://" in d:
                url = d
            elif ":" in d:
                (schema, rest) = d.split(":", 1)
                if schema in self._schema_shortcuts:
                    url = self._schema_shortcuts[schema] % rest
                else:
                    return None
            else:
                url = self._schema_shortcuts["github"] % d

        else:
            return None

        return URIRef(url)

    def normalize_bugs(self, d):
        """https://docs.npmjs.com/cli/v11/configuring-npm/package-json#bugs

        >>> NpmMapping().normalize_bugs({
        ...     'url': 'https://example.org/bugs/',
        ...     'email': 'bugs@example.org'
        ... })
        rdflib.term.URIRef('https://example.org/bugs/')
        >>> NpmMapping().normalize_bugs(
        ...     'https://example.org/bugs/')
        rdflib.term.URIRef('https://example.org/bugs/')
        """
        if isinstance(d, dict) and isinstance(d.get("url"), str):
            url = d["url"]
        elif isinstance(d, str):
            url = d
        else:
            url = ""

        return URIRef(url)

    _parse_author = re.compile(
        r"^ *" r"(?P<name>.*?)" r"( +<(?P<email>.*)>)?" r"( +\((?P<url>.*)\))?" r" *$"
    )

    def translate_author(self, graph: Graph, root, d):
        r"""https://docs.npmjs.com/cli/v11/configuring-npm/package-json#people-fields-author-contributors'

        >>> from pprint import pprint
        >>> root = URIRef("http://example.org/test-software")
        >>> graph = Graph()
        >>> NpmMapping().translate_author(graph, root, {
        ...     'name': 'John Doe',
        ...     'email': 'john.doe@example.org',
        ...     'url': 'https://example.org/~john.doe',
        ... })
        >>> prettyprint_graph(graph, root)
        {
            "@id": ...,
            "http://schema.org/author": {
                "@list": [
                    {
                        "@type": "http://schema.org/Person",
                        "http://schema.org/email": "john.doe@example.org",
                        "http://schema.org/name": "John Doe",
                        "http://schema.org/url": {
                            "@id": "https://example.org/~john.doe"
                        }
                    }
                ]
            }
        }
        >>> graph = Graph()
        >>> NpmMapping().translate_author(graph, root,
        ...     'John Doe <john.doe@example.org> (https://example.org/~john.doe)'
        ... )
        >>> prettyprint_graph(graph, root)
        {
            "@id": ...,
            "http://schema.org/author": {
                "@list": [
                    {
                        "@type": "http://schema.org/Person",
                        "http://schema.org/email": "john.doe@example.org",
                        "http://schema.org/name": "John Doe",
                        "http://schema.org/url": {
                            "@id": "https://example.org/~john.doe"
                        }
                    }
                ]
            }
        }
        >>> graph = Graph()
        >>> NpmMapping().translate_author(graph, root, {
        ...     'name': 'John Doe',
        ...     'email': 'john.doe@example.org',
        ...     'url': 'https:\\\\example.invalid/~john.doe',
        ... })
        >>> prettyprint_graph(graph, root)
        {
            "@id": ...,
            "http://schema.org/author": {
                "@list": [
                    {
                        "@type": "http://schema.org/Person",
                        "http://schema.org/email": "john.doe@example.org",
                        "http://schema.org/name": "John Doe"
                    }
                ]
            }
        }
        """  # noqa
        author = BNode()
        graph.add((author, RDF.type, SCHEMA.Person))
        if isinstance(d, dict):
            name = d.get("name", None)
            email = d.get("email", None)
            url = d.get("url", None)
        elif isinstance(d, str):
            match = self._parse_author.match(d)
            if not match:
                return None
            name = match.group("name")
            email = match.group("email")
            url = match.group("url")
        else:
            return None

        if name and isinstance(name, str):
            graph.add((author, SCHEMA.name, Literal(name)))
        if email and isinstance(email, str):
            graph.add((author, SCHEMA.email, Literal(email)))
        add_url_if_valid(graph, author, SCHEMA.url, url)

        add_list(graph, root, SCHEMA.author, [author])

    def normalize_description(self, description):
        r"""Try to re-decode ``description`` as UTF-16, as this is a somewhat common
        mistake that causes issues in the database because of null bytes in JSON.

        >>> NpmMapping().normalize_description("foo bar")
        rdflib.term.Literal('foo bar')
        >>> NpmMapping().normalize_description(
        ...     "\ufffd\ufffd#\x00 \x00f\x00o\x00o\x00 \x00b\x00a\x00r\x00\r\x00 \x00"
        ... )
        rdflib.term.Literal('foo bar')
        >>> NpmMapping().normalize_description(
        ...     "\ufffd\ufffd\x00#\x00 \x00f\x00o\x00o\x00 \x00b\x00a\x00r\x00\r\x00 "
        ... )
        rdflib.term.Literal('foo bar')
        >>> NpmMapping().normalize_description(
        ...     # invalid UTF-16 and meaningless UTF-8:
        ...     "\ufffd\ufffd\x00#\x00\x00\x00 \x00\x00\x00\x00f\x00\x00\x00\x00"
        ... ) is None
        True
        >>> NpmMapping().normalize_description(
        ...     # ditto (ut looks like little-endian at first)
        ...     "\ufffd\ufffd#\x00\x00\x00 \x00\x00\x00\x00f\x00\x00\x00\x00\x00"
        ... ) is None
        True
        >>> NpmMapping().normalize_description(None) is None
        True
        """
        if not isinstance(description, str):
            return None
        # XXX: if this function ever need to support more cases, consider
        # switching to https://pypi.org/project/ftfy/ instead of adding more hacks
        if description.startswith("\ufffd\ufffd") and "\x00" in description:
            # 2 unicode replacement characters followed by '# ' encoded as UTF-16
            # is a common mistake, which indicates a README.md was saved as UTF-16,
            # and some NPM tool opened it as UTF-8 and used the first line as
            # description.

            description_bytes = description.encode()

            # Strip the the two unicode replacement characters
            assert description_bytes.startswith(b"\xef\xbf\xbd\xef\xbf\xbd")
            description_bytes = description_bytes[6:]

            # If the following attempts fail to recover the description, discard it
            # entirely because the current indexer storage backend (postgresql) cannot
            # store zero bytes in JSON columns.
            description = None

            if not description_bytes.startswith(b"\x00"):
                # try UTF-16 little-endian (the most common) first
                try:
                    description = description_bytes.decode("utf-16le")
                except UnicodeDecodeError:
                    pass
            if description is None:
                # if it fails, try UTF-16 big-endian
                try:
                    description = description_bytes.decode("utf-16be")
                except UnicodeDecodeError:
                    pass

            if description:
                if description.startswith("# "):
                    description = description[2:]
                return Literal(description.rstrip())
            else:
                return None
        return Literal(description)

    def normalize_license(self, s):
        """https://docs.npmjs.com/cli/v11/configuring-npm/package-json#license

        >>> NpmMapping().normalize_license('MIT')
        rdflib.term.URIRef('https://spdx.org/licenses/MIT')
        """
        if isinstance(s, str):
            if s.startswith("SEE LICENSE IN "):
                # Very common pattern, because it is an example in the specification.
                # It is followed by the filename; and the indexer architecture currently
                # does not allow accessing that from metadata mappings.
                # (Plus, an hypothetical license mapping would eventually pick it up)
                return

            # Remove parentheses from the string
            s = s.replace("(", "").replace(")", "")

            if " " in s:
                # Either an SPDX expression, or unusable data
                # Check for SPDX expression first
                # Extract the SPDX expression if it contains OR,
                # ignore licenses with AND or WITH operator.
                if " OR " in s and " AND " not in s and " WITH " not in s:
                    # Multiple licenses, or a license exception
                    # return multiple licenses in a list
                    return [self.normalize_license(x) for x in s.split(" OR ")]

                return
            return SPDX + s

    def normalize_keywords(self, lst):
        """https://docs.npmjs.com/cli/v11/configuring-npm/package-json#homepage

        >>> NpmMapping().normalize_keywords(['foo', 'bar'])
        [rdflib.term.Literal('foo'), rdflib.term.Literal('bar')]
        """
        if isinstance(lst, list):
            return [Literal(x) for x in lst if isinstance(x, str)]
